"""
Fallback directory information.

This file is auto-generated by scripts/update_authorities.py
Do not edit manually.

Fallback directories are stable relays that can be used for
bootstrapping when directory authorities are unreachable.
"""

# pylint: disable=too-many-lines

import random
from dataclasses import dataclass


@dataclass
class FallbackDirectory:
    """Fallback directory information."""

    ip: str
    orport: int
    fingerprint: str  # hex fingerprint
    nickname: str | None = None
    ipv6_address: str | None = None

    @property
    def address(self) -> str:
        """Get formatted address string."""
        return f"{self.ip}:{self.orport}"


FALLBACK_DIRECTORIES = [
    FallbackDirectory(
        ip="45.155.249.124",
        orport=38017,
        fingerprint="4BCB5FC1BE126EF5302447472930A9034070FB08",
        nickname="ZhiliTORNode",
    ),
    FallbackDirectory(
        ip="124.198.131.138",
        orport=80,
        fingerprint="47B56BBC6573EA37A886736145EBB5D6F3EC7480",
        nickname="PlayInTraffic",
    ),
    FallbackDirectory(
        ip="185.235.146.29",
        orport=443,
        fingerprint="5A6976353F19DAB6EB63A2CB95C00039B7D4064A",
        nickname="mwittig1",
        ipv6_address="[2a10:8a40:f313::29]:443",
    ),
    FallbackDirectory(
        ip="81.7.18.7",
        orport=9001,
        fingerprint="0C475BA4D3AA3C289B716F95954CAD616E50C4E5",
        nickname="Freebird32",
        ipv6_address="[2a02:180:6:1::2eee]:9001",
    ),
    FallbackDirectory(
        ip="188.214.88.33",
        orport=9001,
        fingerprint="BC5D9898817E3F07CDD4292B35DA6819368710B7",
        nickname="9u32bg93b2ug9",
    ),
    FallbackDirectory(
        ip="96.9.98.210",
        orport=443,
        fingerprint="9D98766220B32CBD882C1B799D20BE71C3BEAAF6",
        nickname="bigwalkdog",
    ),
    FallbackDirectory(
        ip="45.95.169.109",
        orport=7430,
        fingerprint="7C9FC7C99D7F83980E621BA8D4E0D40FCE53104F",
        nickname="Quetzalcoatl",
        ipv6_address="[2a0e:bfc0:0:14::24a8]:7430",
    ),
    FallbackDirectory(
        ip="192.42.116.214",
        orport=9000,
        fingerprint="31BA70D5332FE49C500FBFFDB0CC5B9D3803CFBF",
        nickname="NTH17R1",
        ipv6_address="[2001:67c:e60:c0c:192:42:116:214]:9000",
    ),
    FallbackDirectory(
        ip="91.208.75.153",
        orport=443,
        fingerprint="5952434A60B6881FFFAF5C3B1C08682FAF5885F5",
        nickname="Nobutomo",
    ),
    FallbackDirectory(
        ip="64.65.1.112",
        orport=443,
        fingerprint="56912E60E4E3003A277B9E46C7EFBC5E01C14C18",
        nickname="nlelamar",
    ),
    FallbackDirectory(
        ip="185.40.4.101",
        orport=13443,
        fingerprint="45FA08CC2A41FE183E4FF5DEDB3AAB13C01FAD39",
        nickname="bauruine",
        ipv6_address="[2a0e:4005:1002:ffff:185:40:4:101]:13443",
    ),
    FallbackDirectory(
        ip="64.65.0.7",
        orport=443,
        fingerprint="862F828CC223FF9916ADFCCEC47BC5E9839413BA",
        nickname="bodegabamz",
    ),
    FallbackDirectory(
        ip="64.65.62.202",
        orport=443,
        fingerprint="032E78EDF6708A3C70ABCC2AED9FAABDB7DB4C43",
        nickname="headieone",
    ),
    FallbackDirectory(
        ip="96.9.98.166",
        orport=443,
        fingerprint="6EF104C3EF66FA71669EE123CC05FEAEFB83A0A5",
        nickname="toosi",
    ),
    FallbackDirectory(
        ip="64.65.1.120",
        orport=443,
        fingerprint="3B0536A960098436EBF9C533711F0BD2C809364F",
        nickname="bigboogie",
    ),
    FallbackDirectory(
        ip="176.65.149.65",
        orport=443,
        fingerprint="3F8B4D9442218B3EA6CC8FD05E37E62B3E6AB361",
        nickname="Famy",
        ipv6_address="[2a05:b0c7:6acd::1]:443",
    ),
    FallbackDirectory(
        ip="45.138.16.107",
        orport=7100,
        fingerprint="64D90959E50D593D80ACD0E29A2BE9051EB66EEA",
        nickname="Quetzalcoatl",
        ipv6_address="[2a12:a800:2:1:45:138:16:107]:7100",
    ),
    FallbackDirectory(
        ip="185.220.101.41",
        orport=10041,
        fingerprint="20209021CE8B8732F8EE57D313A36FF8DC823E5E",
        nickname="ForPrivacyNET",
        ipv6_address="[2a0b:f4c2:2::41]:10041",
    ),
    FallbackDirectory(
        ip="64.65.62.175",
        orport=443,
        fingerprint="CF64EF901AA72D9024B81DBA12DACC6BDF47608A",
        nickname="chinoxl",
    ),
    FallbackDirectory(
        ip="175.114.146.233",
        orport=9001,
        fingerprint="C51D9928A523C3218C1AF09C9450CEA93955E033",
        nickname="LEEJAEMYUNG",
    ),
    FallbackDirectory(
        ip="37.120.26.229",
        orport=9001,
        fingerprint="F864BDEB66C8A11E209456846499C2F995435797",
        nickname="torushopyard",
    ),
    FallbackDirectory(
        ip="45.137.70.158",
        orport=9001,
        fingerprint="B2E1BB031C2882A45E78D9A25A2D039A945378F5",
        nickname="Pjotrek",
        ipv6_address="[2a0e:97c0:3ea:882::1]:9001",
    ),
    FallbackDirectory(
        ip="213.239.215.221",
        orport=9001,
        fingerprint="021DDCD687FD505B7EA7E75DA6CEF0D778AA07FB",
        nickname="GrmmlDag",
        ipv6_address="[2a01:4f8:a0:90b0:789d:6a4:95f3:a78c]:9001",
    ),
    FallbackDirectory(
        ip="101.182.13.197",
        orport=9001,
        fingerprint="8541905725D75B320FACF0C588BA9562FE13BA09",
        nickname="datboigivesweb",
    ),
    FallbackDirectory(
        ip="194.164.245.205",
        orport=443,
        fingerprint="E4C6E31D745BE31F9E150F3BEEE19235D9629FC9",
        nickname="tetris",
        ipv6_address="[2a13:d207:0:40e::1]:443",
    ),
    FallbackDirectory(
        ip="207.153.6.185",
        orport=9001,
        fingerprint="9F8E87DD4B529B58608594AB9233C698BF7B1DEE",
        nickname="DefiantDave",
    ),
    FallbackDirectory(
        ip="194.34.132.169",
        orport=9001,
        fingerprint="6CB2C00C8374D6968698A57940245249FC5B1769",
        nickname="code",
    ),
    FallbackDirectory(
        ip="192.42.116.188",
        orport=22,
        fingerprint="7D9221E99110ED3A1A7E0086D68F920EF1D1DB8C",
        nickname="NTH36R1",
        ipv6_address="[2001:67c:e60:c0c:192:42:116:188]:22",
    ),
    FallbackDirectory(
        ip="45.11.59.28",
        orport=443,
        fingerprint="2FF8FFB28EAA635E291D623E03F231AC265F28F9",
        nickname="AnneMarie",
    ),
    FallbackDirectory(
        ip="192.42.116.189",
        orport=9006,
        fingerprint="5378983BC20642C22BD3A4DB04445F759EBC09E3",
        nickname="NTH37R7",
        ipv6_address="[2001:67c:e60:c0c:192:42:116:189]:9006",
    ),
    FallbackDirectory(
        ip="172.105.20.12",
        orport=443,
        fingerprint="D16D19ED2A8570242708BDABFDFCD7B9E484C8B0",
        nickname="linuxenbyEXIT4",
        ipv6_address="[2600:3c04::f03c:95ff:fe41:d42c]:443",
    ),
    FallbackDirectory(
        ip="188.165.136.211",
        orport=8082,
        fingerprint="061273FD35AD52B31E087E60BBE9F9DA1E05C1A2",
        nickname="Patr",
    ),
    FallbackDirectory(
        ip="64.65.1.42",
        orport=443,
        fingerprint="CE683BCD381A6FBA85B21CF39082C7A79A43DEF8",
        nickname="fabod4l",
    ),
    FallbackDirectory(
        ip="216.73.159.101",
        orport=9001,
        fingerprint="FF03BE4E8F9E7E41E2C7D9682CE4D6496FEE0621",
        nickname="DareDevil",
    ),
    FallbackDirectory(
        ip="162.55.107.247",
        orport=9001,
        fingerprint="675CFAC38BE3C9A26C3A2DD7CBC0E616F68624CA",
        nickname="mullbinde5",
        ipv6_address="[2a01:4f8:2191:3003:3::8]:9001",
    ),
    FallbackDirectory(
        ip="152.53.115.49",
        orport=9001,
        fingerprint="20B31CF54C093348A2D427DD1EBA26EFF998FB78",
        nickname="TorZurFreiheit1",
        ipv6_address="[2a0a:4cc0:80:47a3::2]:443",
    ),
    FallbackDirectory(
        ip="82.165.21.136",
        orport=9001,
        fingerprint="024BEAF1A4B3F01BB7DCC36402EC8764816129D5",
        nickname="mindbyte",
    ),
    FallbackDirectory(
        ip="185.82.126.57",
        orport=443,
        fingerprint="FFAF2849EED3BBB66031173551CFD5E39ECEA324",
        nickname="wfaegfyFDRTlhdtr",
    ),
    FallbackDirectory(
        ip="146.190.224.96",
        orport=9001,
        fingerprint="D70CADC1A398D9ACC28909C09B160DAEF59384CF",
        nickname="slarty2",
        ipv6_address="[2a03:b0c0:2:d0::14cb:6001]:9001",
    ),
    FallbackDirectory(
        ip="102.211.56.200",
        orport=9003,
        fingerprint="C616C915A423BE7DAD47C233E199746CB886CF4E",
        nickname="obzgs5tbmn4q",
    ),
    FallbackDirectory(
        ip="51.68.181.54",
        orport=9004,
        fingerprint="575B47977496E830DF5334F1A2266F16E028CB4B",
        nickname="QuackyForPrivacy",
    ),
    FallbackDirectory(
        ip="154.53.58.161",
        orport=7430,
        fingerprint="BAB2CDAE807EFC4710485F78E8ED359F22C67909",
        nickname="Quetzalcoatl",
        ipv6_address="[2605:a141:2261:4410::1]:7430",
    ),
    FallbackDirectory(
        ip="96.9.98.26",
        orport=443,
        fingerprint="D9B82D372FCA37F82058292FB44E229DE89EE0BE",
        nickname="peeweelongway",
    ),
    FallbackDirectory(
        ip="217.160.252.208",
        orport=9001,
        fingerprint="2E0973F2D6D62064A481F8031265F143210F3883",
        nickname="Wurzelrelay",
        ipv6_address="[2a01:239:28e:6100::1]:9001",
    ),
    FallbackDirectory(
        ip="64.65.0.33",
        orport=443,
        fingerprint="7EA091DF4CE4256ED35DC113D4921867CBF03A0E",
        nickname="megaran",
    ),
    FallbackDirectory(
        ip="23.129.64.145",
        orport=443,
        fingerprint="97A7D10BB2B886695BA766966DE897B5E8B74FB1",
        nickname="mudge",
        ipv6_address="[2620:18c:0:192::e0:145]:443",
    ),
    FallbackDirectory(
        ip="84.165.26.146",
        orport=9001,
        fingerprint="F5165CF26965A27C90B62B016141C60E3E6DD9F5",
        nickname="Unnamed",
    ),
    FallbackDirectory(
        ip="45.80.158.167",
        orport=9000,
        fingerprint="1133B20A95F051F65310C8A32DF92D04751F590C",
        nickname="Quetzalcoatl",
        ipv6_address="[2a12:a800:2:1:45:80:158:167]:9000",
    ),
    FallbackDirectory(
        ip="94.16.115.121",
        orport=9001,
        fingerprint="936AAC50D69C9A3E3B37BA258156547F000C8D9D",
        nickname="LillyCL",
        ipv6_address="[2a03:4000:28:1e8::69]:9001",
    ),
    FallbackDirectory(
        ip="213.183.48.165",
        orport=443,
        fingerprint="4640014F4FD49FF14362C10A9C164A4392F4203B",
        nickname="swordsman",
    ),
    FallbackDirectory(
        ip="162.157.210.69",
        orport=9001,
        fingerprint="8997D74842F8104470418E8FCA7CF8D91F831ED7",
        nickname="Unnamed",
        ipv6_address="[2001:56a:fe40:da00::1337]:9001",
    ),
    FallbackDirectory(
        ip="107.189.1.9",
        orport=9100,
        fingerprint="43EF5141A4417D0BF1B0AB2AD2C3D74C22355D0A",
        nickname="Quetzalcoatl",
        ipv6_address="[2605:6400:30:f920:f523:28ef:3617:8b86]:9100",
    ),
    FallbackDirectory(
        ip="188.245.58.189",
        orport=9001,
        fingerprint="C988DA5E50FB7AF4B76AF66F19574E1921EE2A90",
        nickname="theBNrelay",
    ),
    FallbackDirectory(
        ip="209.114.126.206",
        orport=8080,
        fingerprint="4D2B729D92D699337A99F9A57530F538FD66F1F7",
        nickname="boxoffice206",
    ),
    FallbackDirectory(
        ip="185.241.208.71",
        orport=7100,
        fingerprint="10C079B536D3F682F8A1BB882DED61182646D167",
        nickname="Quetzalcoatl",
        ipv6_address="[2a12:a800:2:1:185:241:208:71]:7100",
    ),
    FallbackDirectory(
        ip="45.83.104.137",
        orport=9100,
        fingerprint="359C5231AC2452D365B64A23C27817A1DFEE56B4",
        nickname="Aramis",
        ipv6_address="[2a03:4000:46:197:b434:d3ff:fe68:d9e1]:9100",
    ),
    FallbackDirectory(
        ip="104.244.79.50",
        orport=9100,
        fingerprint="C2808CAC0896D0AC4EFAAD4ECD467FE245CCFDF7",
        nickname="Quetzalcoatl",
        ipv6_address="[2605:6400:30:f08c:45b7:4a99:a154:694e]:9100",
    ),
    FallbackDirectory(
        ip="185.120.16.176",
        orport=9001,
        fingerprint="79D1CBF51A1FDD140E1733672EDF94C4C1E9C57D",
        nickname="Yuanshen",
    ),
    FallbackDirectory(
        ip="151.244.165.204",
        orport=9001,
        fingerprint="FF3769447DBD2010FE7F31AEC85D29B2B0A5D019",
        nickname="ThePartyBarn",
    ),
    FallbackDirectory(
        ip="173.255.228.134",
        orport=110,
        fingerprint="4ED33415952FA52DB95411A1DA86663EC1B109B9",
        nickname="bedrock",
        ipv6_address="[2600:3c03::f03c:91ff:fedf:407b]:110",
    ),
    FallbackDirectory(
        ip="96.9.98.17",
        orport=443,
        fingerprint="3A85C6698F4655A2127A07F19E8D6B8A489DB79C",
        nickname="jayworthy",
    ),
    FallbackDirectory(
        ip="64.65.1.100",
        orport=443,
        fingerprint="B7169F603A1FC8527C63A514010DE9B2590F093E",
        nickname="kennymuney",
    ),
    FallbackDirectory(
        ip="87.229.85.164",
        orport=58111,
        fingerprint="927BEE8EEC30D1456052972380861889A659CA21",
        nickname="WeaselWave",
    ),
    FallbackDirectory(
        ip="74.108.36.251",
        orport=9001,
        fingerprint="192B3306D8A064510CE6363D5AFFB563E789D246",
        nickname="LittleIsland",
    ),
    FallbackDirectory(
        ip="94.143.137.213",
        orport=443,
        fingerprint="CE4C88B2CA700655AC0CC98F093973EA86CAEE4D",
        nickname="versed03",
    ),
    FallbackDirectory(
        ip="147.135.129.138",
        orport=9500,
        fingerprint="69ADDE95173724077934D6932973E9BB7A40A3BB",
        nickname="prsv",
    ),
    FallbackDirectory(
        ip="194.14.0.243",
        orport=443,
        fingerprint="F7CC37F98E25E75EBC97BFD4FA7955053A0370BC",
        nickname="SrsTubesRelay02",
    ),
    FallbackDirectory(
        ip="81.174.250.136",
        orport=9001,
        fingerprint="87E4FAF25BE28DBD1FC1C9751F2F2D30694A815A",
        nickname="bik",
    ),
    FallbackDirectory(
        ip="172.105.6.34",
        orport=9001,
        fingerprint="D7F59CFB13A93E4C0BF867374B66547E226A7255",
        nickname="AndroidFighter",
        ipv6_address="[2600:3c04::f03c:93ff:fece:e1d1]:9050",
    ),
    FallbackDirectory(
        ip="92.60.38.94",
        orport=9001,
        fingerprint="DD0AA66DD9E4E71FFAFAB658DF8300F1CEEA0364",
        nickname="flowjob04",
        ipv6_address="[2a03:4000:33:1c9:68d7:eff:fe33:3593]:9001",
    ),
    FallbackDirectory(
        ip="195.47.238.178",
        orport=993,
        fingerprint="4BAE2F16CA87CDF0C7E32A11C6A1770EF30AC430",
        nickname="noack14i1",
        ipv6_address="[2a0c:5700:3133:4000:be24:11ff:fe3c:6727]:993",
    ),
    FallbackDirectory(
        ip="107.152.41.240",
        orport=9001,
        fingerprint="2FC20EF67366177AD0F207BFB425209C69A1C5ED",
        nickname="r1tordigi",
        ipv6_address="[2607:9000:0:35::3c1:df81]:9001",
    ),
    FallbackDirectory(
        ip="162.220.14.54",
        orport=9001,
        fingerprint="59F4DCD985864DFF877D5C9887C484F767CD098D",
        nickname="Cassanova",
        ipv6_address="[2604:a00:101:4:216:3eff:fe2f:1c84]:9001",
    ),
    FallbackDirectory(
        ip="194.15.115.91",
        orport=9001,
        fingerprint="5CC73850BA19AF390D7A57BBF2C1EF4BD3290BCD",
        nickname="WaterElementX",
    ),
    FallbackDirectory(
        ip="185.40.4.101",
        orport=8443,
        fingerprint="7A6C085A12D90C56C17A9D1A77D9D1D9B9E592D4",
        nickname="bauruine",
        ipv6_address="[2a0e:4005:1002:ffff:185:40:4:101]:8443",
    ),
    FallbackDirectory(
        ip="64.65.0.4",
        orport=443,
        fingerprint="79361CDC196241A74E50F74A164E8FEFB77C3EA2",
        nickname="shabaamsahdeeq",
    ),
    FallbackDirectory(
        ip="194.55.12.148",
        orport=9001,
        fingerprint="72270EB58EDEBE727AA29E67417628DBCE889FAE",
        nickname="giesskanne",
        ipv6_address="[2a03:4000:31:40:c48f:36ff:fec9:eaf6]:9001",
    ),
    FallbackDirectory(
        ip="51.77.132.82",
        orport=9600,
        fingerprint="1D96ECEECD5C62EA18FE95B61B5936447FA1178D",
        nickname="prsv",
        ipv6_address="[2001:41d0:303:8f52::1]:9600",
    ),
    FallbackDirectory(
        ip="166.84.6.10",
        orport=9001,
        fingerprint="D1F2155B38C21DAC4850C7F702E2946DBD8F5A1D",
        nickname="Unnamed",
        ipv6_address="[2602:f977:800:0:e076:63ff:fe06:2d0a]:9001",
    ),
    FallbackDirectory(
        ip="66.165.241.228",
        orport=9002,
        fingerprint="B2D0823FE5F9047C292B0412603F4801BCCABD07",
        nickname="skibble",
        ipv6_address="[2604:4500:8:14::4]:9002",
    ),
    FallbackDirectory(
        ip="45.95.169.110",
        orport=7430,
        fingerprint="88FA8CBAD03FD602D27985F0DF34FA5DAD9E7BB2",
        nickname="Quetzalcoatl",
        ipv6_address="[2a0e:bfc0:0:2::8a62]:7430",
    ),
    FallbackDirectory(
        ip="37.114.53.201",
        orport=9001,
        fingerprint="2F81B29080727EFF0505E2F66CBD6C1EF6F9F4A9",
        nickname="Ebert",
    ),
    FallbackDirectory(
        ip="64.65.1.169",
        orport=443,
        fingerprint="66653D72465EEA4DD9FC7EE5F09599C482897061",
        nickname="paulolondra",
    ),
    FallbackDirectory(
        ip="148.251.85.195",
        orport=9030,
        fingerprint="77A3ADC5D455778B53C2803761916DFB7DA0A790",
        nickname="redvader",
        ipv6_address="[2a01:4f8:202:6096::2]:9030",
    ),
    FallbackDirectory(
        ip="216.197.207.48",
        orport=9001,
        fingerprint="C366391E67A29CE959A15193D0CF2F9ED94F3578",
        nickname="jorcanada",
    ),
    FallbackDirectory(
        ip="185.195.71.244",
        orport=8443,
        fingerprint="AD86CD1A49573D52A7B6F4A35750F161AAD89C88",
        nickname="DigiGesTor1e2",
    ),
    FallbackDirectory(
        ip="23.92.19.230",
        orport=9001,
        fingerprint="2F67A940520D1C9E66497D3847F0FDEE545AE7C0",
        nickname="fpuzhpx",
        ipv6_address="[2600:3c03::f03c:91ff:fedb:19ed]:9001",
    ),
    FallbackDirectory(
        ip="124.198.132.172",
        orport=7100,
        fingerprint="77313C32AA48AB658299DA039DF6C627C126F570",
        nickname="Quetzalcoatl",
        ipv6_address="[2a12:a800:11:1:124:198:132:172]:7100",
    ),
    FallbackDirectory(
        ip="83.228.199.230",
        orport=8443,
        fingerprint="6DF7E0D81F7A6CF573A27E35D942158063D497CC",
        nickname="Porridge",
        ipv6_address="[2001:1600:13:101::1c87]:8443",
    ),
    FallbackDirectory(
        ip="206.206.192.178",
        orport=9001,
        fingerprint="0C18DD2F789333989A70FDB94C1A3D0E6F8F4081",
        nickname="MiddleEarth",
    ),
    FallbackDirectory(
        ip="23.129.64.206",
        orport=443,
        fingerprint="F6399DB724C3F868833F38D701F0E311AB048BA8",
        nickname="eo206",
        ipv6_address="[2620:18c:0:192::e0:206]:443",
    ),
    FallbackDirectory(
        ip="142.93.228.59",
        orport=443,
        fingerprint="438731B8EFEDB1E592C8034934A55F532DEABAA9",
        nickname="WarheadRelay",
        ipv6_address="[2a03:b0c0:2:f0::1bc:9001]:443",
    ),
    FallbackDirectory(
        ip="71.19.146.15",
        orport=443,
        fingerprint="10F8A0A61A80ABDCD8BFBE9053F8FB2E0C507F71",
        nickname="flyDamFlowertrial",
        ipv6_address="[2605:2700:0:3:a800:ff:fe10:f0d3]:443",
    ),
    FallbackDirectory(
        ip="102.130.113.30",
        orport=9001,
        fingerprint="A4B2F2A51F73D408530FA6C8AF63A7FAFFDAA0FC",
        nickname="Lickitung",
    ),
    FallbackDirectory(
        ip="37.120.194.126",
        orport=443,
        fingerprint="90BBB469D7BD7E4BF80DB90BE2C01995C71D51EE",
        nickname="zarratrident",
    ),
    FallbackDirectory(
        ip="87.106.82.130",
        orport=9001,
        fingerprint="5B99771EB24027E47C04A6260492720C35CA1EDA",
        nickname="ADoorToTor",
    ),
    FallbackDirectory(
        ip="185.242.87.96",
        orport=9001,
        fingerprint="F4F5963A32DE2DB5C5F06261D86BEAED2A0587F9",
        nickname="RMC2",
    ),
    FallbackDirectory(
        ip="64.65.1.70",
        orport=443,
        fingerprint="798809E4A78F31D65235DC0E7EA7C02776E1437D",
        nickname="550madoff",
    ),
    FallbackDirectory(
        ip="171.25.193.78",
        orport=443,
        fingerprint="FDE127305CC410265E66BA82FB5A3E801BFD184E",
        nickname="DFRI17",
        ipv6_address="[2001:67c:289c:4::78]:443",
    ),
    FallbackDirectory(
        ip="45.141.215.56",
        orport=7430,
        fingerprint="719A124624725B23B8CE953BDA53860973EE10D8",
        nickname="Quetzalcoatl",
        ipv6_address="[2a12:a800:2:1:45:141:215:56]:7430",
    ),
    FallbackDirectory(
        ip="64.65.1.33",
        orport=443,
        fingerprint="13A5BCAB9D08209FFE1A860FB4FEEF5B53D8C424",
        nickname="bankrollhayden",
    ),
    FallbackDirectory(
        ip="51.158.146.152",
        orport=443,
        fingerprint="55580D71B317A072F4A4DCF6EA4EDB015734AEE7",
        nickname="VeilsOfTheOnion",
        ipv6_address="[2001:bc8:6010:213:208:a2ff:fe0c:8128]:443",
    ),
    FallbackDirectory(
        ip="80.94.92.92",
        orport=9700,
        fingerprint="27F6AD0910BD90228C5D182F1F9BA199BAE54249",
        nickname="prsv",
    ),
    FallbackDirectory(
        ip="217.255.34.123",
        orport=9001,
        fingerprint="188673CF3937442517301800F383BE53D4A17732",
        nickname="TOPAS",
        ipv6_address="[2003:fd:5709:f700:e65f:1ff:fe00:96c5]:9001",
    ),
    FallbackDirectory(
        ip="45.9.156.105",
        orport=143,
        fingerprint="B9D49260B2EB896E0F7154523C1D7EE988A3CF5F",
        nickname="Quetzalcoatl",
        ipv6_address="[2a0e:bfc7::216:3cff:fe98:45cb]:143",
    ),
    FallbackDirectory(
        ip="185.165.171.84",
        orport=9001,
        fingerprint="4061C553CA88021B8302F0814365070AAE617270",
        nickname="TorExitRomania",
        ipv6_address="[2a06:1700:0:50::11]:9001",
    ),
    FallbackDirectory(
        ip="185.40.4.92",
        orport=13443,
        fingerprint="CF9A87A139A0D03AE7784F7D2DB4207665832A35",
        nickname="bauruine",
        ipv6_address="[2a0e:4005:1002:ffff:185:40:4:92]:13443",
    ),
    FallbackDirectory(
        ip="124.198.132.237",
        orport=9100,
        fingerprint="E86D16D2A0E3503E69971DE999989757C6931195",
        nickname="Quetzalcoatl",
        ipv6_address="[2a12:a800:11:1:124:198:132:237]:9100",
    ),
    FallbackDirectory(
        ip="87.106.135.148",
        orport=443,
        fingerprint="1499D3B81B967C47F18AAC9C1C59576E8CF4F01D",
        nickname="PommespanzerDE5",
        ipv6_address="[2a01:239:319:100::1]:443",
    ),
    FallbackDirectory(
        ip="185.238.129.9",
        orport=9001,
        fingerprint="4E8CE6F5651E7342C1E7E5ED031E82078134FB0D",
        nickname="ymkeo",
        ipv6_address="[2a10:3781:116:1:ff:60ff:fe1e:4473]:9001",
    ),
    FallbackDirectory(
        ip="45.66.35.20",
        orport=1984,
        fingerprint="82E2708957A41EDA0819EE1C5FFC3E2DC38B9FE9",
        nickname="torexitams00",
        ipv6_address="[2a09:61c0::20]:1984",
    ),
    FallbackDirectory(
        ip="5.9.156.17",
        orport=443,
        fingerprint="AFFD147DBCC065A5CEF7258BC1367CC35855A431",
        nickname="Yahta4ee",
    ),
    FallbackDirectory(
        ip="193.189.100.204",
        orport=443,
        fingerprint="327DF526A04129C8AA55B5F3551BC46AC4BC7A6D",
        nickname="TORKeFFORG11",
        ipv6_address="[2a0f:df00:0:255::204]:443",
    ),
    FallbackDirectory(
        ip="158.69.252.116",
        orport=587,
        fingerprint="47A08224924CAAFEB8B8D81066ED007F4E68F2E1",
        nickname="vaahteranlehti",
        ipv6_address="[2607:5300:120:c74::1]:587",
    ),
    FallbackDirectory(
        ip="82.165.171.134",
        orport=443,
        fingerprint="BBC050C12E6A00D36384A23CEA901068F626B615",
        nickname="9g4net4",
        ipv6_address="[2001:ba0:24f:9a00::1]:443",
    ),
    FallbackDirectory(
        ip="94.130.142.182",
        orport=8443,
        fingerprint="50B7C4F6C18A3E5FC1D4E2ECE6B7B316B29FB26C",
        nickname="ElonTusk",
        ipv6_address="[2a01:4f8:13b:2dab::2]:8443",
    ),
    FallbackDirectory(
        ip="84.58.67.136",
        orport=9001,
        fingerprint="EA719AFF7AD3D20464EA4D41C297426C0D1B2071",
        nickname="SSTOR",
    ),
    FallbackDirectory(
        ip="64.65.1.63",
        orport=443,
        fingerprint="C9561E5C0709389715491346D8A22666247F90F8",
        nickname="scootywesson",
    ),
    FallbackDirectory(
        ip="84.226.186.33",
        orport=6969,
        fingerprint="AB40FD78D8A4DE530FB1E0355E74D80367784619",
        nickname="sig9eva",
        ipv6_address="[2001:1715:4e2b:a210:a3a:88ff:fe6e:d46b]:6969",
    ),
    FallbackDirectory(
        ip="23.105.174.243",
        orport=443,
        fingerprint="3BFB3C1C19B5E40D3889093DBD39000FE1D66FF3",
        nickname="BambooRat",
    ),
    FallbackDirectory(
        ip="5.196.131.170",
        orport=9001,
        fingerprint="4C8302C3A6C80081C12BC5A8F9A38917BD923FBE",
        nickname="Unnamed",
    ),
    FallbackDirectory(
        ip="67.186.0.207",
        orport=543,
        fingerprint="EF97A17973B7072C0E3CB2013A3EBAC42346FE42",
        nickname="JAJTorBits3",
        ipv6_address="[2601:547:c601:14a3::1000]:543",
    ),
    FallbackDirectory(
        ip="205.185.123.93",
        orport=9001,
        fingerprint="9C24A441000E6636D4041FFFD669C8DD16672CBC",
        nickname="clan",
    ),
    FallbackDirectory(
        ip="64.62.233.211",
        orport=443,
        fingerprint="32D25F437EB58E246D30AABBEB5D74B5034CB875",
        nickname="tonyd2",
    ),
    FallbackDirectory(
        ip="205.185.119.222",
        orport=443,
        fingerprint="6C620C35C0EA16657960882DE88B28968EBC8689",
        nickname="Cerberus",
        ipv6_address="[2605:6400:20:ae6:e6c6:d90:fee8:5ad5]:443",
    ),
    FallbackDirectory(
        ip="64.65.1.126",
        orport=443,
        fingerprint="53146879222C65C25F3C51002DC0CDF56B184F51",
        nickname="beaniesigel",
    ),
    FallbackDirectory(
        ip="212.51.149.67",
        orport=9001,
        fingerprint="E42DEAC28AB38BD89312FB92DA1B5D0CDC09386C",
        nickname="tetris",
    ),
    FallbackDirectory(
        ip="157.90.77.166",
        orport=9001,
        fingerprint="02A0D8DDF4DA4E601C1C66143BB0C8A0F2F3C857",
        nickname="changeme",
    ),
    FallbackDirectory(
        ip="185.220.101.36",
        orport=10136,
        fingerprint="C32905D18F2B0A104B039A10ED41C3D0C0BDF03A",
        nickname="ForPrivacyNET",
        ipv6_address="[2a0b:f4c2:2::36]:10136",
    ),
    FallbackDirectory(
        ip="45.14.233.204",
        orport=443,
        fingerprint="104AE1991C61426D9D3256758928B4CE6EF2AFAB",
        nickname="mikrogravitation4",
        ipv6_address="[2a0e:1580:1000:0:be24:11ff:fe60:2797]:443",
    ),
    FallbackDirectory(
        ip="91.219.236.101",
        orport=443,
        fingerprint="54FAC042C6F2FDEB449DD5B9EDE6833D155C538B",
        nickname="Lyle",
        ipv6_address="[2a10:c800:1:9cc6::1]:443",
    ),
    FallbackDirectory(
        ip="64.62.233.206",
        orport=9000,
        fingerprint="0833FA63031FD1007AAF386576B6CA2452460C5B",
        nickname="50cent3",
    ),
    FallbackDirectory(
        ip="140.238.97.211",
        orport=9001,
        fingerprint="0B53297F0BD687103A00A514E7D2AB83C70F3700",
        nickname="ruffletor",
    ),
    FallbackDirectory(
        ip="23.92.34.118",
        orport=443,
        fingerprint="3018498D945665829287BD9F59E7C8FFBA51830D",
        nickname="lauraschlessinger",
    ),
    FallbackDirectory(
        ip="185.107.57.64",
        orport=443,
        fingerprint="737BBB37E406B02EF8E38A54C3AB8AC2143069C9",
        nickname="eddy",
    ),
    FallbackDirectory(
        ip="64.62.233.204",
        orport=8443,
        fingerprint="272B6E6276D7AC4D63A9495461B5D7053B046E8E",
        nickname="kanyewest5",
    ),
    FallbackDirectory(
        ip="188.68.43.192",
        orport=9001,
        fingerprint="98867F5242505A2DD581C85B1CE256090B87F138",
        nickname="Piratenpartei03",
        ipv6_address="[2a03:4000:17:688:b854:7ff:fe48:bdcb]:9001",
    ),
    FallbackDirectory(
        ip="217.12.202.85",
        orport=443,
        fingerprint="E1DA0F90454AED535ACE9B415F78C1A1359E29A1",
        nickname="lyingbush",
        ipv6_address="[2a02:27aa::88]:443",
    ),
    FallbackDirectory(
        ip="179.43.158.176",
        orport=443,
        fingerprint="698B8708CF4EA02D27CA09274F3351B4F66A5834",
        nickname="greenlantern",
    ),
    FallbackDirectory(
        ip="160.119.249.24",
        orport=443,
        fingerprint="801EF80845848E685C4534A2408D32D5443187DF",
        nickname="montrose",
    ),
    FallbackDirectory(
        ip="185.241.208.115",
        orport=9100,
        fingerprint="F74A9BDD0CA2EFE07907D70AA18B43D923E3E745",
        nickname="Quetzalcoatl",
        ipv6_address="[2a12:a800:2:1:185:241:208:115]:9100",
    ),
    FallbackDirectory(
        ip="45.86.86.150",
        orport=9696,
        fingerprint="8764E9BE25CE405C76E1DD0F0937D9BF6EA16A0C",
        nickname="FreedomForEveryone",
    ),
    FallbackDirectory(
        ip="94.142.241.41",
        orport=80,
        fingerprint="F04AC7D177EC79308D6959DBE042A247D9761635",
        nickname="NTH100R2",
    ),
    FallbackDirectory(
        ip="23.129.64.183",
        orport=443,
        fingerprint="8036F4287ED61CE0C42B5734F873EE79C0CC7E27",
        nickname="eo183",
        ipv6_address="[2620:18c:0:192::e0:183]:443",
    ),
    FallbackDirectory(
        ip="185.220.101.96",
        orport=443,
        fingerprint="52CD98902F6376E84DF2415B906F1426D585564D",
        nickname="Digitalcourage4ipaa",
        ipv6_address="[2a0b:f4c2:4::96]:443",
    ),
    FallbackDirectory(
        ip="31.133.0.210",
        orport=443,
        fingerprint="90705D63956EDD5904D173F1998F421C5FC5B6FC",
        nickname="MariePierre",
    ),
    FallbackDirectory(
        ip="94.230.208.148",
        orport=443,
        fingerprint="1BDBE9C0F7034E6789A9BF7BED82BE2045F0F5B7",
        nickname="DigiGesTor3e3",
        ipv6_address="[2a02:418:6017::148]:443",
    ),
    FallbackDirectory(
        ip="194.87.252.100",
        orport=1443,
        fingerprint="0625588B87F580A5B203A2C46020807860740662",
        nickname="Honos",
    ),
    FallbackDirectory(
        ip="109.69.218.176",
        orport=443,
        fingerprint="FA7136FD4F8A727810A22ED2096AF60F872C2F41",
        nickname="vandewoestijne",
    ),
    FallbackDirectory(
        ip="136.35.254.165",
        orport=999,
        fingerprint="DCC59C82C41C1E3E26BC4119D4CC579EF4B58391",
        nickname="AngryUnicorn",
    ),
    FallbackDirectory(
        ip="78.31.67.127",
        orport=9001,
        fingerprint="31A9565D0A51DD94EB81B6459C30276C5D790E20",
        nickname="mullbinde6",
        ipv6_address="[2001:4ba0:ffa4:413:200::208]:9001",
    ),
    FallbackDirectory(
        ip="89.58.5.0",
        orport=853,
        fingerprint="304EF9E012BE51190C6DDD5CFBAD8E3F3E875A43",
        nickname="TorRelay2c",
        ipv6_address="[2a03:4000:5e:c47::1]:853",
    ),
    FallbackDirectory(
        ip="174.160.2.27",
        orport=9001,
        fingerprint="8D7159CAB26F26A81A8A06A3BF5897576965E35E",
        nickname="AlliumAdPortum",
        ipv6_address="[2601:647:6701:5100:73b2:b84b:3c33:a8a1]:9001",
    ),
    FallbackDirectory(
        ip="64.65.1.185",
        orport=443,
        fingerprint="1C47BDDCE3618F6C9FC4F90D2D7E2B5FCD40AD6C",
        nickname="playafly",
    ),
    FallbackDirectory(
        ip="140.238.255.68",
        orport=443,
        fingerprint="26BB860977EE6FE3F52F9E7452D04A28F45D6FD9",
        nickname="acabworldwide2",
    ),
    FallbackDirectory(
        ip="192.42.116.201",
        orport=9001,
        fingerprint="4E93A68717B358B594F7ABAA5742D5EFE0E3E0C6",
        nickname="NTH38R2",
        ipv6_address="[2001:67c:e60:c0c:192:42:116:201]:9001",
    ),
    FallbackDirectory(
        ip="84.251.69.149",
        orport=9001,
        fingerprint="18582D11115B9CA4122101A3AF39F30A802B1B4B",
        nickname="zer0bitz",
        ipv6_address="[2001:2003:f745:9500::bad:da7a]:9001",
    ),
    FallbackDirectory(
        ip="64.65.62.17",
        orport=443,
        fingerprint="20B1B5FC996C4F48BF1B615B7472ED498D985DF1",
        nickname="recycledj",
    ),
    FallbackDirectory(
        ip="45.63.15.42",
        orport=443,
        fingerprint="8E9E3239EC9A082F5337C5F45A8E706A86BA19B0",
        nickname="zubrowka",
    ),
    FallbackDirectory(
        ip="23.163.216.135",
        orport=443,
        fingerprint="2683A6EB5F2DAB673CE92858904862EE1CEB4CB7",
        nickname="mkn7787relay0",
        ipv6_address="[2602:ffd1:7:6b::a]:443",
    ),
    FallbackDirectory(
        ip="109.192.104.34",
        orport=443,
        fingerprint="1E6C798C8624780A977EA758A200C0296EC9CA6A",
        nickname="fluppi",
        ipv6_address="[2a02:8071:5f00:31a0:20d:48ff:fe5f:b239]:9001",
    ),
    FallbackDirectory(
        ip="160.119.249.240",
        orport=443,
        fingerprint="8428C02CCD1CF01715C91F1819BA507C1A79A931",
        nickname="dani",
    ),
    FallbackDirectory(
        ip="180.13.1.145",
        orport=4001,
        fingerprint="7435AF094B162CAE8904850F13BD79786BB54D1B",
        nickname="cuccnl10G",
    ),
    FallbackDirectory(
        ip="88.80.26.3",
        orport=443,
        fingerprint="C90A7324D99F311B4AB1C3DE15CF1DDD4C56FF31",
        nickname="PRQseTORexit4",
        ipv6_address="[2a00:16b0:1:243::7012:e817]:443",
    ),
    FallbackDirectory(
        ip="77.20.42.234",
        orport=443,
        fingerprint="6428176DC76120CE8EF980595539130F9D40AC0D",
        nickname="Ponyhof",
    ),
    FallbackDirectory(
        ip="31.184.236.104",
        orport=9001,
        fingerprint="48407E7B3AA2E3F3B7DC7980F2D76479489AF0F5",
        nickname="koob0927",
    ),
    FallbackDirectory(
        ip="64.65.1.8",
        orport=443,
        fingerprint="2AF9C0C8AE27CC0D6694BF139F961CCA6B8C5757",
        nickname="diamondatl",
    ),
    FallbackDirectory(
        ip="145.239.136.129",
        orport=8080,
        fingerprint="D5146BA3FF0C4F3EFAE486BD1F693C2B786AA8AF",
        nickname="Starfleet",
    ),
    FallbackDirectory(
        ip="198.50.191.95",
        orport=443,
        fingerprint="39F096961ED2576975C866D450373A9913AFDC92",
        nickname="shhovh",
    ),
    FallbackDirectory(
        ip="37.114.53.9",
        orport=9001,
        fingerprint="26E3C67C9BB4B9303B01452AF9E71455686D241B",
        nickname="SkibidiCrackhead",
        ipv6_address="[2a12:3e80:400:8::]:9001",
    ),
    FallbackDirectory(
        ip="193.200.229.34",
        orport=1080,
        fingerprint="286F53DD87AEE0B085C34F34259AFC0B09A259FC",
        nickname="kodakblack",
    ),
    FallbackDirectory(
        ip="80.67.172.162",
        orport=9001,
        fingerprint="59F2831A9BEB7AB3D5803445F946EA9FEEEFA217",
        nickname="algrothendieck2",
    ),
    FallbackDirectory(
        ip="50.54.51.20",
        orport=9001,
        fingerprint="C5BB82D7E0491C48B1F1E2E09B98FF1C95C205F5",
        nickname="NICKNET",
    ),
    FallbackDirectory(
        ip="203.31.40.236",
        orport=443,
        fingerprint="D3A55B115FC350DF5A26FFB2C098EB0B30456CAE",
        nickname="Aramis",
    ),
    FallbackDirectory(
        ip="107.175.59.252",
        orport=443,
        fingerprint="0F5A78ECBA449016B00F05A85398D5AD3DC7A895",
        nickname="YoungBrother2",
    ),
    FallbackDirectory(
        ip="5.2.78.126",
        orport=9001,
        fingerprint="E5DEA9F6570207DB9841CA04C240BCDCFEC3200D",
        nickname="Unnamed",
    ),
    FallbackDirectory(
        ip="185.165.169.239",
        orport=443,
        fingerprint="CD94752BFBC94EC82F081B86E95241BB71D7CF99",
        nickname="impishingyserawn",
    ),
    FallbackDirectory(
        ip="45.141.215.238",
        orport=8430,
        fingerprint="34B80D703F4D6350146B684E66D962A23A830117",
        nickname="Quetzalcoatl",
        ipv6_address="[2a12:a800:2:1:45:141:215:238]:8430",
    ),
    FallbackDirectory(
        ip="192.42.116.188",
        orport=8080,
        fingerprint="45B06605BD7AB37BFE0452767706B98559517526",
        nickname="NTH36R7",
        ipv6_address="[2001:67c:e60:c0c:192:42:116:188]:8080",
    ),
    FallbackDirectory(
        ip="82.149.227.236",
        orport=9003,
        fingerprint="67E7B7D4BFCB1E0B4DF2CDB33B1FDE4B69C9DD62",
        nickname="f14480488695e0c7b27",
    ),
    FallbackDirectory(
        ip="64.65.62.242",
        orport=443,
        fingerprint="B7859ED84EFEE8FF8267587580A1A7CCCE2CB82E",
        nickname="jkwon",
    ),
    FallbackDirectory(
        ip="45.90.4.235",
        orport=9001,
        fingerprint="F475E5987E44A2B7B143A2BEE3F8128EEFD7E5A3",
        nickname="0x1ea7deadbeef",
        ipv6_address="[2a03:4000:58:7d:de1:ea7:dead:beef]:9090",
    ),
    FallbackDirectory(
        ip="109.70.100.245",
        orport=80,
        fingerprint="38B0226F4318EEB45D4D6777E18A17B50615F3A5",
        nickname="bender",
        ipv6_address="[2a03:e600:100:c3::10]:80",
    ),
    FallbackDirectory(
        ip="217.160.114.102",
        orport=9001,
        fingerprint="4C122DFAED8BD43AD63BB1045AF53A9708F17695",
        nickname="RelayUli",
    ),
    FallbackDirectory(
        ip="64.65.1.39",
        orport=443,
        fingerprint="A9EA412E9035B285D68EAA994FADFB1EF49324D8",
        nickname="parlaed4l",
    ),
    FallbackDirectory(
        ip="93.95.231.88",
        orport=9001,
        fingerprint="9DFF7F006FA33615DADD4EA775E1CEF73E7D7630",
        nickname="keepingfaith",
    ),
    FallbackDirectory(
        ip="95.217.112.245",
        orport=443,
        fingerprint="A2CD32D9D0668DB764AD68C745CE29693CA851B9",
        nickname="arbitraryTessa0",
        ipv6_address="[2a01:4f9:4a:2e4e::2]:443",
    ),
    FallbackDirectory(
        ip="160.119.253.114",
        orport=443,
        fingerprint="17919460EFE0DFC2B0CF0D6C453AC72B252630F5",
        nickname="lon",
    ),
    FallbackDirectory(
        ip="38.135.24.30",
        orport=9002,
        fingerprint="EC36350A0239EAFB1D5626816EED167123C4B3CD",
        nickname="NeelTorExit1A",
        ipv6_address="[2602:f6f6:2:ad5c::1]:9002",
    ),
    FallbackDirectory(
        ip="103.149.168.242",
        orport=9001,
        fingerprint="E7017A361C44F7403B186E812C48A06916AD5D83",
        nickname="as215248node7",
        ipv6_address="[2a14:38c0:1::242]:9001",
    ),
    FallbackDirectory(
        ip="159.69.11.74",
        orport=1050,
        fingerprint="8CF9D1663EC7FC79BFF31AEAC1C907D075153470",
        nickname="OkurwaRelay",
        ipv6_address="[2a01:4f8:1c1c:31b5::1]:1050",
    ),
    FallbackDirectory(
        ip="51.81.201.207",
        orport=143,
        fingerprint="A9A567CC15293CD3CD6B2BC12A435F37731843A2",
        nickname="Echidna4",
        ipv6_address="[2604:2dc0:202:300::cd0]:143",
    ),
    FallbackDirectory(
        ip="5.104.84.183",
        orport=143,
        fingerprint="F1467AA85A0CDDE645EC690A9602095E63AFF097",
        nickname="Quetzalcoatl",
        ipv6_address="[2400:d320:2261:6860::1]:143",
    ),
    FallbackDirectory(
        ip="185.40.4.149",
        orport=9004,
        fingerprint="E0AE4772051D74DE4DC1A525D9670BC12ED2EDAF",
        nickname="4727",
        ipv6_address="[2a0e:4005:1002:ffff:185:40:4:149]:9004",
    ),
    FallbackDirectory(
        ip="37.114.57.163",
        orport=9001,
        fingerprint="E8B91980497DF39125A8F713135EAE761E6E0897",
        nickname="napster",
    ),
    FallbackDirectory(
        ip="209.58.180.236",
        orport=9001,
        fingerprint="0371E4EC901C4FC77FEA2AA5BE4218F0BE1F6CE3",
        nickname="hong",
    ),
    FallbackDirectory(
        ip="38.135.24.32",
        orport=9002,
        fingerprint="E0711FF93A8655D7CE5485240A0B7AF19CE512C8",
        nickname="NeelTorExit3A",
        ipv6_address="[2602:f6f6:2:d31a::1]:9002",
    ),
    FallbackDirectory(
        ip="51.89.23.41",
        orport=9000,
        fingerprint="D0EDEFA81312D510D61666904120A762681A82A1",
        nickname="prsv",
    ),
    FallbackDirectory(
        ip="5.45.104.176",
        orport=9000,
        fingerprint="AA4644F0EC589EEA2F501BB867E32E599F8169DA",
        nickname="Quetzalcoatl",
        ipv6_address="[2a03:4000:6:102b:c457:f3ff:feb0:a6d0]:9000",
    ),
]


def get_fallbacks() -> list[FallbackDirectory]:
    """Get list of fallback directories."""
    return FALLBACK_DIRECTORIES


def get_random_fallback() -> FallbackDirectory:
    """Get a random fallback directory."""
    return random.choice(FALLBACK_DIRECTORIES)


def get_shuffled_fallbacks() -> list[FallbackDirectory]:
    """Get all fallback directories in random order."""
    fallbacks = list(FALLBACK_DIRECTORIES)
    random.shuffle(fallbacks)
    return fallbacks



"""Tests for Tor cell format implementation."""

from torscope.onion.cell import (
    CELL_LEN_V3,
    CELL_LEN_V4,
    AuthChallengeCell,
    Cell,
    CellCommand,
    CertsCell,
    NetInfoCell,
    VersionsCell,
)


class TestCellCommand:
    """Tests for CellCommand enum."""

    def test_fixed_length_commands(self):
        """Test that fixed-length commands are correctly identified."""
        assert not CellCommand.is_variable_length(CellCommand.PADDING)
        assert not CellCommand.is_variable_length(CellCommand.CREATE)
        assert not CellCommand.is_variable_length(CellCommand.RELAY)
        assert not CellCommand.is_variable_length(CellCommand.DESTROY)
        assert not CellCommand.is_variable_length(CellCommand.NETINFO)

    def test_variable_length_commands(self):
        """Test that variable-length commands are correctly identified."""
        # VERSIONS is special - variable length despite being < 128
        assert CellCommand.is_variable_length(CellCommand.VERSIONS)
        # Commands >= 128 are variable length
        assert CellCommand.is_variable_length(CellCommand.VPADDING)
        assert CellCommand.is_variable_length(CellCommand.CERTS)
        assert CellCommand.is_variable_length(CellCommand.AUTH_CHALLENGE)

    def test_command_values(self):
        """Test that command values match Tor specification."""
        assert CellCommand.PADDING == 0
        assert CellCommand.CREATE == 1
        assert CellCommand.RELAY == 3
        assert CellCommand.DESTROY == 4
        assert CellCommand.VERSIONS == 7
        assert CellCommand.NETINFO == 8
        assert CellCommand.CERTS == 129
        assert CellCommand.AUTH_CHALLENGE == 130


class TestCell:
    """Tests for base Cell class."""

    def test_pack_fixed_cell_v4(self):
        """Test packing a fixed-length cell for link protocol 4."""
        cell = Cell(circ_id=0, command=CellCommand.PADDING, payload=b"")
        packed = cell.pack(link_protocol=4)
        assert len(packed) == CELL_LEN_V4
        # CircID (4 bytes) + Command (1 byte) + padding
        assert packed[:4] == b"\x00\x00\x00\x00"  # CircID = 0
        assert packed[4] == CellCommand.PADDING

    def test_pack_fixed_cell_v3(self):
        """Test packing a fixed-length cell for link protocol 3."""
        cell = Cell(circ_id=0, command=CellCommand.PADDING, payload=b"")
        packed = cell.pack(link_protocol=3)
        assert len(packed) == CELL_LEN_V3
        # CircID (2 bytes) + Command (1 byte) + padding
        assert packed[:2] == b"\x00\x00"  # CircID = 0
        assert packed[2] == CellCommand.PADDING

    def test_pack_variable_cell_v4(self):
        """Test packing a variable-length cell for link protocol 4."""
        payload = b"test payload"
        cell = Cell(circ_id=0, command=CellCommand.VERSIONS, payload=payload)
        packed = cell.pack(link_protocol=4)
        # CircID (4 bytes) + Command (1 byte) + Length (2 bytes) + payload
        assert len(packed) == 4 + 1 + 2 + len(payload)
        assert packed[:4] == b"\x00\x00\x00\x00"  # CircID = 0
        assert packed[4] == CellCommand.VERSIONS
        assert packed[5:7] == b"\x00\x0c"  # Length = 12
        assert packed[7:] == payload

    def test_unpack_fixed_cell_v4(self):
        """Test unpacking a fixed-length cell for link protocol 4."""
        # Create a NETINFO cell
        packed = b"\x00\x00\x00\x00" + bytes([CellCommand.NETINFO]) + b"\x00" * (CELL_LEN_V4 - 5)
        cell = Cell.unpack(packed, link_protocol=4)
        assert cell.circ_id == 0
        assert cell.command == CellCommand.NETINFO

    def test_unpack_fixed_cell_v3(self):
        """Test unpacking a fixed-length cell for link protocol 3."""
        packed = b"\x00\x00" + bytes([CellCommand.NETINFO]) + b"\x00" * (CELL_LEN_V3 - 3)
        cell = Cell.unpack(packed, link_protocol=3)
        assert cell.circ_id == 0
        assert cell.command == CellCommand.NETINFO

    def test_roundtrip_fixed_cell(self):
        """Test that packing and unpacking a fixed cell preserves data."""
        payload = b"test data"
        cell = Cell(circ_id=123, command=CellCommand.PADDING, payload=payload)
        packed = cell.pack(link_protocol=4)
        unpacked = Cell.unpack(packed, link_protocol=4)
        assert unpacked.circ_id == cell.circ_id
        assert unpacked.command == cell.command

    def test_roundtrip_variable_cell(self):
        """Test that packing and unpacking a variable cell preserves data."""
        payload = b"test data"
        cell = Cell(circ_id=0, command=CellCommand.VPADDING, payload=payload)
        packed = cell.pack(link_protocol=4)
        unpacked = Cell.unpack(packed, link_protocol=4)
        assert unpacked.circ_id == cell.circ_id
        assert unpacked.command == cell.command
        assert unpacked.payload == payload


class TestVersionsCell:
    """Tests for VERSIONS cell."""

    def test_pack_versions_cell(self):
        """Test packing a VERSIONS cell."""
        cell = VersionsCell(versions=[4, 5])
        packed = cell.pack()
        # VERSIONS always uses 2-byte CircID
        # CircID (2 bytes) + Command (1 byte) + Length (2 bytes) + versions (2 bytes each)
        assert len(packed) == 2 + 1 + 2 + 4
        assert packed[:2] == b"\x00\x00"  # CircID = 0
        assert packed[2] == CellCommand.VERSIONS
        assert packed[3:5] == b"\x00\x04"  # Length = 4

    def test_unpack_versions_cell(self):
        """Test unpacking a VERSIONS cell."""
        # CircID (2 bytes) + Command + Length + versions [3, 4, 5]
        packed = b"\x00\x00\x07\x00\x06\x00\x03\x00\x04\x00\x05"
        cell = VersionsCell.unpack(packed)
        assert cell.versions == [3, 4, 5]

    def test_roundtrip_versions_cell(self):
        """Test packing and unpacking a VERSIONS cell."""
        original = VersionsCell(versions=[3, 4, 5])
        packed = original.pack()
        unpacked = VersionsCell.unpack(packed)
        assert unpacked.versions == original.versions


class TestNetInfoCell:
    """Tests for NETINFO cell."""

    def test_pack_netinfo_cell(self):
        """Test packing a NETINFO cell."""
        cell = NetInfoCell(
            timestamp=1234567890,
            other_address=(4, b"\x7f\x00\x00\x01"),  # 127.0.0.1
            my_addresses=[],
        )
        packed = cell.pack(link_protocol=4)
        # Should be a fixed-length cell
        assert len(packed) == CELL_LEN_V4

    def test_unpack_netinfo_cell(self):
        """Test unpacking a NETINFO cell."""
        # Create a simple NETINFO cell
        cell = NetInfoCell(
            timestamp=1234567890,
            other_address=(4, b"\x7f\x00\x00\x01"),
            my_addresses=[(4, b"\xc0\xa8\x01\x01")],  # 192.168.1.1
        )
        packed = cell.pack(link_protocol=4)
        unpacked = NetInfoCell.unpack(packed, link_protocol=4)
        assert unpacked.timestamp == 1234567890
        assert unpacked.other_address == (4, b"\x7f\x00\x00\x01")
        assert len(unpacked.my_addresses) == 1


class TestCertsCell:
    """Tests for CERTS cell."""

    def test_pack_certs_cell(self):
        """Test packing a CERTS cell."""
        cell = CertsCell(
            certificates=[
                (1, b"certificate1"),
                (2, b"certificate2"),
            ]
        )
        packed = cell.pack(link_protocol=4)
        # Variable-length cell
        # CircID (4) + Command (1) + Length (2) + NumCerts (1) + certs
        assert packed[4] == CellCommand.CERTS

    def test_unpack_certs_cell(self):
        """Test unpacking a CERTS cell."""
        cell = CertsCell(
            certificates=[
                (1, b"cert1"),
                (2, b"cert2"),
            ]
        )
        packed = cell.pack(link_protocol=4)
        unpacked = CertsCell.unpack(packed, link_protocol=4)
        assert len(unpacked.certificates) == 2
        assert unpacked.certificates[0] == (1, b"cert1")
        assert unpacked.certificates[1] == (2, b"cert2")


class TestAuthChallengeCell:
    """Tests for AUTH_CHALLENGE cell."""

    def test_pack_auth_challenge_cell(self):
        """Test packing an AUTH_CHALLENGE cell."""
        cell = AuthChallengeCell(
            challenge=b"\x00" * 32,
            methods=[1, 3],
        )
        packed = cell.pack(link_protocol=4)
        assert packed[4] == CellCommand.AUTH_CHALLENGE

    def test_unpack_auth_challenge_cell(self):
        """Test unpacking an AUTH_CHALLENGE cell."""
        challenge = bytes(range(32))
        cell = AuthChallengeCell(challenge=challenge, methods=[1, 3])
        packed = cell.pack(link_protocol=4)
        unpacked = AuthChallengeCell.unpack(packed, link_protocol=4)
        assert unpacked.challenge == challenge
        assert unpacked.methods == [1, 3]



"""Tests for cryptographic utilities."""

import base64

import pytest
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa

from torscope.crypto import (
    compute_rsa_key_fingerprint,
    extract_signed_portion,
    load_rsa_public_key,
    verify_consensus_signature,
    verify_rsa_signature,
)


# Generate a test RSA key pair for testing
def generate_test_keypair() -> tuple[rsa.RSAPrivateKey, str]:
    """Generate a test RSA key pair."""
    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
    )
    public_key = private_key.public_key()
    pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.PKCS1,
    ).decode("utf-8")
    return private_key, pem


class TestLoadRsaPublicKey:
    """Tests for load_rsa_public_key function."""

    def test_load_valid_pem_key(self) -> None:
        """Test loading a valid PEM-encoded RSA public key."""
        _, pem = generate_test_keypair()
        key = load_rsa_public_key(pem)
        assert key is not None

    def test_load_key_without_headers(self) -> None:
        """Test loading a key without PEM headers."""
        _, pem = generate_test_keypair()
        # Remove headers
        lines = pem.strip().split("\n")
        base64_only = "\n".join(lines[1:-1])
        key = load_rsa_public_key(base64_only)
        assert key is not None

    def test_load_invalid_key(self) -> None:
        """Test loading invalid key data raises ValueError."""
        with pytest.raises(ValueError):
            load_rsa_public_key("not a valid key")


class TestComputeRsaKeyFingerprint:
    """Tests for compute_rsa_key_fingerprint function."""

    def test_fingerprint_format(self) -> None:
        """Test that fingerprint is uppercase hex."""
        _, pem = generate_test_keypair()
        fingerprint = compute_rsa_key_fingerprint(pem)
        assert fingerprint.isupper()
        assert all(c in "0123456789ABCDEF" for c in fingerprint)
        assert len(fingerprint) == 40  # SHA1 = 20 bytes = 40 hex chars

    def test_same_key_same_fingerprint(self) -> None:
        """Test that same key produces same fingerprint."""
        _, pem = generate_test_keypair()
        fp1 = compute_rsa_key_fingerprint(pem)
        fp2 = compute_rsa_key_fingerprint(pem)
        assert fp1 == fp2

    def test_different_keys_different_fingerprints(self) -> None:
        """Test that different keys produce different fingerprints."""
        _, pem1 = generate_test_keypair()
        _, pem2 = generate_test_keypair()
        fp1 = compute_rsa_key_fingerprint(pem1)
        fp2 = compute_rsa_key_fingerprint(pem2)
        assert fp1 != fp2


def create_tor_signature(
    private_key: rsa.RSAPrivateKey, data: bytes, algorithm: str = "sha1"
) -> bytes:
    """Create a signature in Tor's format (PKCS#1 v1.5 with raw hash, no DigestInfo OID)."""
    import hashlib

    # Compute the hash
    if algorithm == "sha256":
        hash_bytes = hashlib.sha256(data).digest()
    else:
        hash_bytes = hashlib.sha1(data).digest()

    # Get key size
    key_size_bytes = private_key.key_size // 8

    # Build PKCS#1 v1.5 padded message: 00 01 [FF...] 00 [hash]
    # Calculate padding length
    padding_len = key_size_bytes - 3 - len(hash_bytes)  # 3 = 00 01 ... 00

    # Build the padded message
    padded = bytes([0, 1]) + (b"\xff" * padding_len) + bytes([0]) + hash_bytes

    # RSA sign (raw RSA operation)
    padded_int = int.from_bytes(padded, "big")
    numbers = private_key.private_numbers()
    signature_int = pow(padded_int, numbers.d, numbers.public_numbers.n)
    signature = signature_int.to_bytes(key_size_bytes, "big")

    return signature


class TestVerifyRsaSignature:
    """Tests for verify_rsa_signature function.

    Note: Tor uses a non-standard PKCS#1 v1.5 format with raw hashes
    (no DigestInfo OID), so we need to create signatures in that format.
    """

    def test_valid_signature_sha1(self) -> None:
        """Test verifying a valid SHA1 signature in Tor format."""
        private_key, pem = generate_test_keypair()
        data = b"test data to sign"

        # Create signature in Tor's format (raw hash, no DigestInfo OID)
        signature = create_tor_signature(private_key, data, "sha1")

        # Verify
        public_key = load_rsa_public_key(pem)
        assert verify_rsa_signature(public_key, signature, data, "sha1") is True

    def test_valid_signature_sha256(self) -> None:
        """Test verifying a valid SHA256 signature in Tor format."""
        private_key, pem = generate_test_keypair()
        data = b"test data to sign"

        # Create signature in Tor's format (raw hash, no DigestInfo OID)
        signature = create_tor_signature(private_key, data, "sha256")

        # Verify
        public_key = load_rsa_public_key(pem)
        assert verify_rsa_signature(public_key, signature, data, "sha256") is True

    def test_invalid_signature(self) -> None:
        """Test that invalid signature returns False."""
        _, pem = generate_test_keypair()
        public_key = load_rsa_public_key(pem)
        data = b"test data"
        bad_signature = b"not a valid signature"

        assert verify_rsa_signature(public_key, bad_signature, data, "sha1") is False

    def test_wrong_data(self) -> None:
        """Test that signature doesn't verify with wrong data."""
        private_key, pem = generate_test_keypair()
        data = b"original data"
        wrong_data = b"different data"

        signature = create_tor_signature(private_key, data, "sha1")

        public_key = load_rsa_public_key(pem)
        assert verify_rsa_signature(public_key, signature, wrong_data, "sha1") is False


class TestVerifyConsensusSignature:
    """Tests for verify_consensus_signature function."""

    def test_valid_signature_with_pem_headers(self) -> None:
        """Test verifying signature with PEM-style headers."""
        private_key, pem = generate_test_keypair()
        data = b"network-status-version 3\ntest data\n"

        # Create signature in Tor's format
        signature = create_tor_signature(private_key, data, "sha1")

        sig_b64 = (
            "-----BEGIN SIGNATURE-----\n"
            + base64.b64encode(signature).decode()
            + "\n-----END SIGNATURE-----"
        )

        assert verify_consensus_signature(pem, sig_b64, data, "sha1") is True

    def test_invalid_key(self) -> None:
        """Test that invalid key returns False."""
        assert verify_consensus_signature("bad key", "sig", b"data", "sha1") is False


class TestExtractSignedPortion:
    """Tests for extract_signed_portion function.

    Note: The signed portion is the same for all signatures in a consensus document.
    The identity parameter is kept for API compatibility but is ignored.
    According to Tor spec, the signed portion is from 'network-status-version'
    through the space after 'directory-signature' (sha1) or
    'directory-signature sha256' (sha256).
    """

    def test_extract_simple_document(self) -> None:
        """Test extracting signed portion from simple document."""
        doc = """network-status-version 3
vote-status consensus
directory-signature ABC123 DEF456
-----BEGIN SIGNATURE-----
dGVzdA==
-----END SIGNATURE-----
"""
        signed = extract_signed_portion(doc, "ABC123", "sha1")
        assert signed is not None
        assert signed.startswith(b"network-status-version")
        # The signed portion ends at "directory-signature " (with space)
        assert signed.endswith(b"directory-signature ")

    def test_extract_with_algorithm(self) -> None:
        """Test extracting with algorithm in signature line.

        The signed portion always ends at 'directory-signature ' regardless
        of what algorithm is specified in the signature line. The algorithm
        only indicates which hash to use on the signed portion.
        """
        doc = """network-status-version 3
vote-status consensus
directory-signature sha256 ABC123 DEF456
-----BEGIN SIGNATURE-----
dGVzdA==
-----END SIGNATURE-----
"""
        signed = extract_signed_portion(doc, "ABC123", "sha256")
        assert signed is not None
        # The signed portion ends at "directory-signature " (with space)
        # NOT at "directory-signature sha256 " - the algorithm is part of the
        # unsigned portion
        assert signed.endswith(b"directory-signature ")

    def test_extract_multiple_signatures(self) -> None:
        """Test extracting when multiple signatures exist.

        The signed portion uses the FIRST occurrence of the end marker,
        regardless of which identity is specified.
        """
        doc = """network-status-version 3
vote-status consensus
directory-signature FIRST111 KEY111
-----BEGIN SIGNATURE-----
c2lnMQ==
-----END SIGNATURE-----
directory-signature SECOND22 KEY222
-----BEGIN SIGNATURE-----
c2lnMg==
-----END SIGNATURE-----
"""
        # Both identities should return the same signed portion
        # (ending at the FIRST "directory-signature ")
        signed1 = extract_signed_portion(doc, "FIRST111", "sha1")
        signed2 = extract_signed_portion(doc, "SECOND22", "sha1")
        assert signed1 is not None
        assert signed2 is not None
        assert signed1 == signed2
        assert signed1.endswith(b"directory-signature ")

    def test_extract_identity_ignored(self) -> None:
        """Test that identity parameter is ignored."""
        doc = """network-status-version 3
directory-signature ABC123 DEF456
"""
        # Identity is ignored, so any value (including non-existent) works
        signed = extract_signed_portion(doc, "NOTFOUND", "sha1")
        assert signed is not None
        assert signed.endswith(b"directory-signature ")

    def test_extract_no_version(self) -> None:
        """Test that missing network-status-version returns None."""
        doc = """vote-status consensus
directory-signature ABC123 DEF456
"""
        signed = extract_signed_portion(doc, "ABC123", "sha1")
        assert signed is None

    def test_extract_no_signature(self) -> None:
        """Test that missing directory-signature returns None."""
        doc = """network-status-version 3
vote-status consensus
"""
        signed = extract_signed_portion(doc, "ABC123", "sha1")
        assert signed is None



"""
obfs4 handshake implementation.

The obfs4 handshake establishes an encrypted session between client and server.
It uses an ntor-like key exchange with Elligator2-encoded public keys.

Protocol overview:
1. Client sends: X' (representative) || padding || mark || MAC
2. Server sends: Y' (representative) || auth || padding || mark || MAC
3. Both derive session keys from shared secrets

Reference: https://github.com/Yawning/obfs4/blob/master/doc/obfs4-spec.txt
"""

from __future__ import annotations

import base64
import hashlib
import hmac
import os
import time
from dataclasses import dataclass

from cryptography.hazmat.primitives.asymmetric.x25519 import (
    X25519PrivateKey,
    X25519PublicKey,
)

from torscope.onion.obfs4.elligator import elligator2_decode, elligator2_encode

# Protocol constants
NODE_ID_LENGTH = 20
PUBLIC_KEY_LENGTH = 32
REPRESENTATIVE_LENGTH = 32
AUTH_LENGTH = 32
MARK_LENGTH = 16
MAC_LENGTH = 16

# Padding ranges
CLIENT_MIN_PAD_LENGTH = 85  # Minimum padding
CLIENT_MAX_PAD_LENGTH = 8192 - REPRESENTATIVE_LENGTH - MARK_LENGTH - MAC_LENGTH

# Maximum handshake size
MAX_HANDSHAKE_LENGTH = 8192

# Key derivation constants
PROTO_ID = b"ntor-curve25519-sha256-1:obfs4"
T_MAC = PROTO_ID + b":mac"
T_KEY = PROTO_ID + b":key_extract"
T_VERIFY = PROTO_ID + b":verify"
M_EXPAND = PROTO_ID + b":key_expand"
SERVER_TIMING_TAG = b"Server"


class HandshakeError(Exception):
    """Error during obfs4 handshake."""


@dataclass
class Obfs4ServerCert:
    """
    Parsed obfs4 server certificate.

    The cert parameter in bridge lines encodes:
    - 20 bytes: Node ID
    - 32 bytes: Curve25519 public key

    Format: base64(node_id || public_key) with padding stripped
    """

    node_id: bytes  # 20 bytes
    public_key: bytes  # 32 bytes (Curve25519)

    @classmethod
    def from_string(cls, cert_str: str) -> Obfs4ServerCert:
        """
        Parse cert parameter from bridge line.

        The cert uses a compact base64 encoding with padding stripped.
        We need to add padding before decoding.
        """
        # Add padding characters
        # Standard base64 needs length to be multiple of 4
        padding_needed = (4 - len(cert_str) % 4) % 4
        padded = cert_str + "=" * padding_needed

        try:
            raw = base64.b64decode(padded)
        except Exception as e:
            raise HandshakeError(f"Invalid cert base64: {e}") from e

        if len(raw) != NODE_ID_LENGTH + PUBLIC_KEY_LENGTH:
            raise HandshakeError(
                f"Invalid cert length: expected {NODE_ID_LENGTH + PUBLIC_KEY_LENGTH}, "
                f"got {len(raw)}"
            )

        return cls(
            node_id=raw[:NODE_ID_LENGTH],
            public_key=raw[NODE_ID_LENGTH:],
        )


def _hmac_sha256(key: bytes, data: bytes) -> bytes:
    """Compute HMAC-SHA256."""
    return hmac.new(key, data, hashlib.sha256).digest()


def _hmac_sha256_128(key: bytes, data: bytes) -> bytes:
    """Compute HMAC-SHA256 truncated to 128 bits (16 bytes)."""
    return _hmac_sha256(key, data)[:16]


def _hkdf_expand(key_seed: bytes, info: bytes, length: int) -> bytes:
    """
    HKDF-Expand as per RFC 5869.

    This matches Tor's key expansion algorithm.
    """
    result = b""
    prev = b""
    counter = 1

    while len(result) < length:
        prev = _hmac_sha256(key_seed, prev + info + bytes([counter]))
        result += prev
        counter += 1

    return result[:length]


def _get_epoch_hours() -> int:
    """Get current Unix epoch in hours."""
    return int(time.time()) // 3600


@dataclass
class ClientHandshake:
    """
    Client-side obfs4 handshake state.

    Manages the ephemeral keypair and handshake message generation.
    """

    server_cert: Obfs4ServerCert
    iat_mode: int = 0

    # Generated during __post_init__
    _private_key: X25519PrivateKey | None = None
    _public_key: bytes | None = None
    _representative: bytes | None = None
    _epoch_hours: int | None = None

    def __post_init__(self) -> None:
        """Generate an Elligator2-encodable ephemeral keypair."""
        self._epoch_hours = _get_epoch_hours()

        # Generate keypair until we get one that's Elligator2-encodable
        while True:
            self._private_key = X25519PrivateKey.generate()
            self._public_key = self._private_key.public_key().public_bytes_raw()
            self._representative = elligator2_encode(self._public_key)
            if self._representative is not None:
                break

    def _make_key_material(self) -> bytes:
        """Create the key material for MAC computation: server_pk || node_id."""
        return self.server_cert.public_key + self.server_cert.node_id

    def generate_request(self, pad_length: int | None = None) -> bytes:
        """
        Generate the client handshake request.

        Format: X' (32) || padding || M_C (16) || MAC_C (16)

        Args:
            pad_length: Padding length (default: random in valid range)

        Returns:
            Client handshake request bytes
        """
        if self._representative is None:
            raise HandshakeError("Handshake not initialized")

        # Determine padding length
        if pad_length is None:
            pad_length = os.urandom(1)[0] % (CLIENT_MAX_PAD_LENGTH - CLIENT_MIN_PAD_LENGTH)
            pad_length += CLIENT_MIN_PAD_LENGTH

        padding = os.urandom(pad_length)

        # Key material for MACs
        key_material = self._make_key_material()

        # M_C = HMAC-SHA256-128(B || ID, X')
        mark = _hmac_sha256_128(key_material, self._representative)

        # Epoch hour as 4 bytes (little-endian)
        epoch_bytes = (self._epoch_hours or 0).to_bytes(4, "little")

        # MAC_C = HMAC-SHA256-128(B || ID, X' || P_C || M_C || E)
        mac_input = self._representative + padding + mark + epoch_bytes
        mac = _hmac_sha256_128(key_material, mac_input)

        return self._representative + padding + mark + mac

    def process_response(self, response: bytes) -> bytes:
        """
        Process server response and derive session keys.

        Server response format: Y' (32) || AUTH (32) || padding || M_S (16) || MAC_S (16)

        Args:
            response: Server handshake response

        Returns:
            144 bytes of key material for framing

        Raises:
            HandshakeError: If response is invalid or authentication fails
        """
        if len(response) < REPRESENTATIVE_LENGTH + AUTH_LENGTH + MARK_LENGTH + MAC_LENGTH:
            raise HandshakeError(f"Response too short: {len(response)} bytes")

        key_material = self._make_key_material()

        # Find the mark in the response
        # The mark should be at: len(response) - MAC_LENGTH - MARK_LENGTH
        # But we need to scan for it since padding length is variable

        # First, try to find mark by scanning
        mark_pos = self._find_mark(response, key_material)
        if mark_pos < 0:
            raise HandshakeError("Could not find server mark in response")

        # Extract server representative and auth
        server_repr = response[:REPRESENTATIVE_LENGTH]
        server_auth = response[REPRESENTATIVE_LENGTH : REPRESENTATIVE_LENGTH + AUTH_LENGTH]

        # Verify MAC
        mac_pos = mark_pos + MARK_LENGTH
        if mac_pos + MAC_LENGTH > len(response):
            raise HandshakeError("Response truncated")

        epoch_bytes = (self._epoch_hours or 0).to_bytes(4, "little")
        expected_mac = _hmac_sha256_128(key_material, response[:mac_pos] + epoch_bytes)
        actual_mac = response[mac_pos : mac_pos + MAC_LENGTH]

        if not hmac.compare_digest(expected_mac, actual_mac):
            raise HandshakeError("Server MAC verification failed")

        # Decode server's public key from representative
        server_pubkey_bytes = elligator2_decode(server_repr)
        server_pubkey = X25519PublicKey.from_public_bytes(server_pubkey_bytes)

        # Compute shared secrets
        # EXP(Y, x) - shared secret with server's ephemeral
        shared_xy = self._private_key.exchange(server_pubkey)  # type: ignore

        # EXP(B, x) - shared secret with server's static key
        server_static = X25519PublicKey.from_public_bytes(self.server_cert.public_key)
        shared_xb = self._private_key.exchange(server_static)  # type: ignore

        # Derive keys
        keys = self._derive_keys(
            shared_xy,
            shared_xb,
            self._public_key or b"",
            server_pubkey_bytes,
            server_auth,
        )

        return keys

    def _find_mark(self, response: bytes, key_material: bytes) -> int:
        """
        Find the server's mark in the response.

        The mark is HMAC-SHA256-128(B || ID, Y') where Y' is the first 32 bytes.
        """
        if len(response) < REPRESENTATIVE_LENGTH:
            return -1

        server_repr = response[:REPRESENTATIVE_LENGTH]
        expected_mark = _hmac_sha256_128(key_material, server_repr)

        # Scan for mark (it should be near the end)
        min_pos = REPRESENTATIVE_LENGTH + AUTH_LENGTH
        max_pos = len(response) - MARK_LENGTH - MAC_LENGTH

        for pos in range(max_pos, min_pos - 1, -1):
            if response[pos : pos + MARK_LENGTH] == expected_mark:
                return pos

        return -1

    def _derive_keys(
        self,
        shared_xy: bytes,
        shared_xb: bytes,
        client_pubkey: bytes,
        server_pubkey: bytes,
        server_auth: bytes,
    ) -> bytes:
        """
        Derive session keys from shared secrets.

        Uses ntor-like key derivation.
        """
        # secret_input = EXP(Y,x) || EXP(B,x) || ID || B || X || Y || PROTO_ID
        secret_input = (
            shared_xy
            + shared_xb
            + self.server_cert.node_id
            + self.server_cert.public_key
            + client_pubkey
            + server_pubkey
            + PROTO_ID
        )

        # Compute verify and expected auth
        key_seed = _hmac_sha256(T_KEY, secret_input)
        verify = _hmac_sha256(T_VERIFY, secret_input)

        # auth = H(verify || ID || B || Y || X || PROTO_ID || "Server", T_MAC)
        auth_input = (
            verify
            + self.server_cert.node_id
            + self.server_cert.public_key
            + server_pubkey
            + client_pubkey
            + PROTO_ID
            + SERVER_TIMING_TAG
        )
        expected_auth = _hmac_sha256(T_MAC, auth_input)

        if not hmac.compare_digest(expected_auth, server_auth):
            raise HandshakeError("Server auth verification failed")

        # Expand keys: 144 bytes total (72 per direction)
        # Each direction: key (32) || nonce_prefix (16) || sip_key (16) || ofb_iv (8)
        return _hkdf_expand(key_seed, M_EXPAND, 144)



"""
Extra-info descriptor parsing.

This module provides functionality to parse Tor extra-info descriptors,
which contain relay statistics like bandwidth history, exit traffic,
and directory request statistics.
"""

# pylint: disable=duplicate-code

import re
from datetime import UTC, datetime

from torscope.directory.models import BandwidthHistory, ExtraInfoDescriptor


class ExtraInfoParser:
    """Parser for Tor extra-info descriptors."""

    @staticmethod
    def parse(content: bytes) -> list[ExtraInfoDescriptor]:
        """
        Parse extra-info descriptors from raw bytes.

        Args:
            content: Raw descriptor bytes (may contain multiple descriptors)

        Returns:
            List of parsed ExtraInfoDescriptor objects
        """
        text = content.decode("utf-8", errors="replace")
        return ExtraInfoParser.parse_text(text)

    @staticmethod
    def parse_text(text: str) -> list[ExtraInfoDescriptor]:
        """
        Parse extra-info descriptors from text.

        Args:
            text: Descriptor text (may contain multiple descriptors)

        Returns:
            List of parsed ExtraInfoDescriptor objects
        """
        descriptors = []

        # Split into individual descriptors
        raw_descriptors = re.split(r"(?=^extra-info\s)", text, flags=re.MULTILINE)

        for raw in raw_descriptors:
            raw = raw.strip()
            if not raw or not raw.startswith("extra-info "):
                continue

            try:
                descriptor = ExtraInfoParser._parse_single(raw)
                if descriptor:
                    descriptors.append(descriptor)
            # pylint: disable-next=broad-exception-caught
            except Exception:
                continue

        return descriptors

    @staticmethod
    def _parse_single(text: str) -> ExtraInfoDescriptor | None:
        """Parse a single extra-info descriptor."""
        lines = text.split("\n")

        # Parse extra-info line (required)
        # extra-info <nickname> <fingerprint>
        if not lines[0].startswith("extra-info "):
            return None

        parts = lines[0].split()
        if len(parts) < 3:
            return None

        nickname = parts[1]
        fingerprint = parts[2].upper()

        # Initialize with defaults
        published = datetime.now(UTC)
        write_history: BandwidthHistory | None = None
        read_history: BandwidthHistory | None = None
        dirreq_write_history: BandwidthHistory | None = None
        dirreq_read_history: BandwidthHistory | None = None
        geoip_db_digest: str | None = None
        geoip6_db_digest: str | None = None
        dirreq_stats_end: datetime | None = None
        dirreq_v3_ips: dict[str, int] = {}
        dirreq_v3_reqs: dict[str, int] = {}
        dirreq_v3_resp: dict[str, int] = {}
        entry_stats_end: datetime | None = None
        entry_ips: dict[str, int] = {}
        exit_stats_end: datetime | None = None
        exit_kibibytes_written: dict[str, int] = {}
        exit_kibibytes_read: dict[str, int] = {}
        exit_streams_opened: dict[str, int] = {}
        cell_stats_end: datetime | None = None
        cell_processed_cells: list[float] = []
        cell_queued_cells: list[float] = []
        cell_time_in_queue: list[int] = []
        hidserv_stats_end: datetime | None = None
        hidserv_rend_relayed_cells: int | None = None
        hidserv_dir_onions_seen: int | None = None

        for line in lines[1:]:
            line = line.strip()

            if line.startswith("published "):
                published = ExtraInfoParser._parse_datetime(line[10:])

            elif line.startswith("write-history "):
                write_history = ExtraInfoParser._parse_history(line[14:])

            elif line.startswith("read-history "):
                read_history = ExtraInfoParser._parse_history(line[13:])

            elif line.startswith("dirreq-write-history "):
                dirreq_write_history = ExtraInfoParser._parse_history(line[21:])

            elif line.startswith("dirreq-read-history "):
                dirreq_read_history = ExtraInfoParser._parse_history(line[20:])

            elif line.startswith("geoip-db-digest "):
                geoip_db_digest = line[16:]

            elif line.startswith("geoip6-db-digest "):
                geoip6_db_digest = line[17:]

            elif line.startswith("dirreq-stats-end "):
                dirreq_stats_end = ExtraInfoParser._parse_stats_end(line[17:])

            elif line.startswith("dirreq-v3-ips "):
                dirreq_v3_ips = ExtraInfoParser._parse_key_value_pairs(line[14:])

            elif line.startswith("dirreq-v3-reqs "):
                dirreq_v3_reqs = ExtraInfoParser._parse_key_value_pairs(line[15:])

            elif line.startswith("dirreq-v3-resp "):
                dirreq_v3_resp = ExtraInfoParser._parse_key_value_pairs(line[15:])

            elif line.startswith("entry-stats-end "):
                entry_stats_end = ExtraInfoParser._parse_stats_end(line[16:])

            elif line.startswith("entry-ips "):
                entry_ips = ExtraInfoParser._parse_key_value_pairs(line[10:])

            elif line.startswith("exit-stats-end "):
                exit_stats_end = ExtraInfoParser._parse_stats_end(line[15:])

            elif line.startswith("exit-kibibytes-written "):
                exit_kibibytes_written = ExtraInfoParser._parse_key_value_pairs(line[23:])

            elif line.startswith("exit-kibibytes-read "):
                exit_kibibytes_read = ExtraInfoParser._parse_key_value_pairs(line[20:])

            elif line.startswith("exit-streams-opened "):
                exit_streams_opened = ExtraInfoParser._parse_key_value_pairs(line[20:])

            elif line.startswith("cell-stats-end "):
                cell_stats_end = ExtraInfoParser._parse_stats_end(line[15:])

            elif line.startswith("cell-processed-cells "):
                cell_processed_cells = ExtraInfoParser._parse_float_list(line[21:])

            elif line.startswith("cell-queued-cells "):
                cell_queued_cells = ExtraInfoParser._parse_float_list(line[18:])

            elif line.startswith("cell-time-in-queue "):
                cell_time_in_queue = ExtraInfoParser._parse_int_list(line[19:])

            elif line.startswith("hidserv-stats-end "):
                hidserv_stats_end = ExtraInfoParser._parse_stats_end(line[18:])

            elif line.startswith("hidserv-rend-relayed-cells "):
                hidserv_rend_relayed_cells = ExtraInfoParser._parse_obfuscated_int(line[27:])

            elif line.startswith("hidserv-dir-onions-seen "):
                hidserv_dir_onions_seen = ExtraInfoParser._parse_obfuscated_int(line[24:])

        return ExtraInfoDescriptor(
            nickname=nickname,
            fingerprint=fingerprint,
            published=published,
            write_history=write_history,
            read_history=read_history,
            dirreq_write_history=dirreq_write_history,
            dirreq_read_history=dirreq_read_history,
            geoip_db_digest=geoip_db_digest,
            geoip6_db_digest=geoip6_db_digest,
            dirreq_stats_end=dirreq_stats_end,
            dirreq_v3_ips=dirreq_v3_ips,
            dirreq_v3_reqs=dirreq_v3_reqs,
            dirreq_v3_resp=dirreq_v3_resp,
            entry_stats_end=entry_stats_end,
            entry_ips=entry_ips,
            exit_stats_end=exit_stats_end,
            exit_kibibytes_written=exit_kibibytes_written,
            exit_kibibytes_read=exit_kibibytes_read,
            exit_streams_opened=exit_streams_opened,
            cell_stats_end=cell_stats_end,
            cell_processed_cells=cell_processed_cells,
            cell_queued_cells=cell_queued_cells,
            cell_time_in_queue=cell_time_in_queue,
            hidserv_stats_end=hidserv_stats_end,
            hidserv_rend_relayed_cells=hidserv_rend_relayed_cells,
            hidserv_dir_onions_seen=hidserv_dir_onions_seen,
            raw_descriptor=text,
        )

    @staticmethod
    def _parse_datetime(dt_str: str) -> datetime:
        """Parse datetime string like '2024-01-15 12:00:00'."""
        try:
            dt = datetime.strptime(dt_str.strip(), "%Y-%m-%d %H:%M:%S")
            return dt.replace(tzinfo=UTC)
        except ValueError:
            return datetime.now(UTC)

    @staticmethod
    def _parse_stats_end(line: str) -> datetime | None:
        """Parse stats-end line like '2024-01-15 12:00:00 (86400 s)'."""
        match = re.match(r"(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2})", line)
        if match:
            return ExtraInfoParser._parse_datetime(match.group(1))
        return None

    @staticmethod
    def _parse_history(line: str) -> BandwidthHistory | None:
        """
        Parse bandwidth history line.

        Format: YYYY-MM-DD HH:MM:SS (NSEC s) NUM,NUM,...
        """
        match = re.match(
            r"(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}) \((\d+) s\)\s*(.*)",
            line,
        )
        if not match:
            return None

        timestamp = ExtraInfoParser._parse_datetime(match.group(1))
        interval_seconds = int(match.group(2))
        values_str = match.group(3)

        values: list[int] = []
        if values_str:
            for v in values_str.split(","):
                try:
                    values.append(int(v))
                except ValueError:
                    pass

        return BandwidthHistory(
            timestamp=timestamp,
            interval_seconds=interval_seconds,
            values=values,
        )

    @staticmethod
    def _parse_key_value_pairs(line: str) -> dict[str, int]:
        """
        Parse key=value pairs like 'us=100,de=50,fr=25'.
        """
        result: dict[str, int] = {}
        if not line.strip():
            return result

        for pair in line.split(","):
            if "=" in pair:
                key, value = pair.split("=", 1)
                try:
                    result[key] = int(value)
                except ValueError:
                    pass
        return result

    @staticmethod
    def _parse_float_list(line: str) -> list[float]:
        """Parse comma-separated float values."""
        result: list[float] = []
        for v in line.split(","):
            try:
                result.append(float(v))
            except ValueError:
                pass
        return result

    @staticmethod
    def _parse_int_list(line: str) -> list[int]:
        """Parse comma-separated int values."""
        result: list[int] = []
        for v in line.split(","):
            try:
                result.append(int(v))
            except ValueError:
                pass
        return result

    @staticmethod
    def _parse_obfuscated_int(line: str) -> int | None:
        """
        Parse obfuscated integer with Laplace noise.

        Format: NUM delta_f=X epsilon=Y bin_size=Z
        We just extract the main number.
        """
        parts = line.split()
        if parts:
            try:
                return int(parts[0])
            except ValueError:
                pass
        return None



"""Tests for server descriptor parsing."""

from torscope.directory.descriptor import ServerDescriptorParser
from torscope.directory.models import ServerDescriptor

# Sample server descriptor for testing
SAMPLE_DESCRIPTOR = b"""router TestRelay 192.0.2.1 9001 0 0
identity-ed25519
-----BEGIN ED25519 CERT-----
AQQABp1bAXe...
-----END ED25519 CERT-----
master-key-ed25519 AbCdEfGhIjKlMnOpQrStUvWxYz0123456789ABCDEFGH
platform Tor 0.4.8.10 on Linux
proto Cons=1-2 Desc=1-2 DirCache=2 HSDir=2 HSIntro=4-5 Link=1-5 Microdesc=1-2 Relay=1-4
published 2024-01-15 12:00:00
fingerprint ABCD 1234 EFGH 5678 IJKL 9012 MNOP 3456 QRST 7890
uptime 86400
bandwidth 1000000 2000000 500000
extra-info-digest ABC123 DEF456
onion-key
-----BEGIN RSA PUBLIC KEY-----
MIGJAoGBALZpAOmKd...
-----END RSA PUBLIC KEY-----
signing-key
-----BEGIN RSA PUBLIC KEY-----
MIGJAoGBAN1k2B3...
-----END RSA PUBLIC KEY-----
ntor-onion-key AbCdEfGhIjKlMnOpQrStUvWxYz0123456789ABCD
contact admin@example.com
family $AAAA $BBBB $CCCC
reject *:25
reject *:119
reject *:135-139
accept *:*
router-sig-ed25519 AbCdE...
router-signature
-----BEGIN SIGNATURE-----
ZWx...
-----END SIGNATURE-----
"""


class TestServerDescriptorParser:
    """Tests for ServerDescriptorParser."""

    def test_parse_basic_descriptor(self):
        """Test parsing a basic server descriptor."""
        descriptors = ServerDescriptorParser.parse(SAMPLE_DESCRIPTOR)
        assert len(descriptors) == 1

        desc = descriptors[0]
        assert desc.nickname == "TestRelay"
        assert desc.ip == "192.0.2.1"
        assert desc.orport == 9001
        assert desc.dirport == 0

    def test_parse_fingerprint(self):
        """Test fingerprint parsing."""
        descriptors = ServerDescriptorParser.parse(SAMPLE_DESCRIPTOR)
        desc = descriptors[0]
        # Fingerprint should be joined without spaces
        assert desc.fingerprint == "ABCD1234EFGH5678IJKL9012MNOP3456QRST7890"

    def test_parse_bandwidth(self):
        """Test bandwidth parsing."""
        descriptors = ServerDescriptorParser.parse(SAMPLE_DESCRIPTOR)
        desc = descriptors[0]
        assert desc.bandwidth_avg == 1000000
        assert desc.bandwidth_burst == 2000000
        assert desc.bandwidth_observed == 500000

    def test_parse_uptime(self):
        """Test uptime parsing."""
        descriptors = ServerDescriptorParser.parse(SAMPLE_DESCRIPTOR)
        desc = descriptors[0]
        assert desc.uptime == 86400
        assert desc.uptime_days == 1.0

    def test_parse_contact(self):
        """Test contact parsing."""
        descriptors = ServerDescriptorParser.parse(SAMPLE_DESCRIPTOR)
        desc = descriptors[0]
        assert desc.contact == "admin@example.com"

    def test_parse_family(self):
        """Test family parsing."""
        descriptors = ServerDescriptorParser.parse(SAMPLE_DESCRIPTOR)
        desc = descriptors[0]
        assert desc.family == ["$AAAA", "$BBBB", "$CCCC"]

    def test_parse_exit_policy(self):
        """Test exit policy parsing."""
        descriptors = ServerDescriptorParser.parse(SAMPLE_DESCRIPTOR)
        desc = descriptors[0]
        assert len(desc.exit_policy) == 4
        assert "reject *:25" in desc.exit_policy
        assert "accept *:*" in desc.exit_policy

    def test_parse_ntor_key(self):
        """Test ntor onion key parsing."""
        descriptors = ServerDescriptorParser.parse(SAMPLE_DESCRIPTOR)
        desc = descriptors[0]
        assert desc.ntor_onion_key == "AbCdEfGhIjKlMnOpQrStUvWxYz0123456789ABCD"

    def test_parse_platform(self):
        """Test platform parsing."""
        descriptors = ServerDescriptorParser.parse(SAMPLE_DESCRIPTOR)
        desc = descriptors[0]
        assert desc.platform == "Tor 0.4.8.10 on Linux"
        assert desc.tor_version == "0.4.8.10"

    def test_parse_protocols(self):
        """Test protocol parsing."""
        descriptors = ServerDescriptorParser.parse(SAMPLE_DESCRIPTOR)
        desc = descriptors[0]
        assert desc.protocols is not None
        assert "Cons" in desc.protocols
        assert desc.protocols["Cons"] == [1, 2]
        assert "Link" in desc.protocols
        assert desc.protocols["Link"] == [1, 2, 3, 4, 5]

    def test_parse_empty(self):
        """Test parsing empty content."""
        descriptors = ServerDescriptorParser.parse(b"")
        assert descriptors == []

    def test_parse_invalid(self):
        """Test parsing invalid content."""
        descriptors = ServerDescriptorParser.parse(b"not a descriptor")
        assert descriptors == []

    def test_bandwidth_mbps_property(self):
        """Test bandwidth_mbps property."""
        desc = ServerDescriptor(
            nickname="test",
            fingerprint="ABC123",
            published=None,  # type: ignore
            ip="192.0.2.1",
            orport=9001,
            bandwidth_observed=5_000_000,
        )
        assert desc.bandwidth_mbps == 5.0

    def test_uptime_days_property_none(self):
        """Test uptime_days when uptime is None."""
        desc = ServerDescriptor(
            nickname="test",
            fingerprint="ABC123",
            published=None,  # type: ignore
            ip="192.0.2.1",
            orport=9001,
        )
        assert desc.uptime_days is None



"""HSDir hashring selection for v3 hidden services.

This module implements the HSDir (Hidden Service Directory) selection
algorithm as specified in rend-spec-v3.txt.

The algorithm places both services and HSDirs on a hashring, and finds
the closest HSDirs to a service's blinded key for descriptor storage/lookup.
"""

from __future__ import annotations

import base64
import struct
from dataclasses import dataclass

from torscope.crypto import sha3_256
from torscope.directory.models import ConsensusDocument, RouterStatusEntry

# Number of replicas for descriptor distribution
HS_N_REPLICAS = 2

# Number of HSDirs to select per replica (spread)
HSDIR_SPREAD_STORE = 4  # For storing
HSDIR_SPREAD_FETCH = 3  # For fetching (we try fewer)

# Time period length in minutes (24 hours)
HS_TIME_PERIOD_LENGTH = 1440


@dataclass
class HSDirectoryEntry:
    """An HSDir with its computed index on the hashring."""

    router: RouterStatusEntry
    index: bytes  # 32-byte SHA3-256 hash (position on ring)
    ed25519_id: bytes | None  # 32-byte Ed25519 identity if available

    def __lt__(self, other: HSDirectoryEntry) -> bool:
        """Compare by index for sorting on the ring."""
        return self.index < other.index


class HSDirectoryRing:
    """Hashring of HSDir relays for hidden service directory lookup.

    The hashring is computed from the consensus and shared random value.
    Each HSDir's position on the ring is determined by:

        hsdir_index = SHA3-256("node-idx" | ed25519_id | shared_random |
                               INT_8(period_num) | INT_8(period_length))

    Services compute their position similarly (replica is 1-indexed: 1, 2):

        hs_index = SHA3-256("store-at-idx" | blinded_key |
                            INT_8(replica) | INT_8(period_length) |
                            INT_8(period_num))
    """

    def __init__(
        self,
        consensus: ConsensusDocument,
        time_period: int,
        use_second_srv: bool = False,
        ed25519_map: dict[str, bytes] | None = None,
    ):
        """Initialize the HSDir hashring.

        Args:
            consensus: Network consensus document
            time_period: Current time period number
            use_second_srv: If True, use shared_rand_previous instead of current
            ed25519_map: Pre-computed map of microdesc_hash -> Ed25519 identity (optional)
        """
        self.consensus = consensus
        self.time_period = time_period
        self.period_length = HS_TIME_PERIOD_LENGTH

        # Get shared random value from consensus
        if use_second_srv and consensus.shared_rand_previous:
            self.shared_random = self._decode_srv(consensus.shared_rand_previous)
        elif consensus.shared_rand_current:
            self.shared_random = self._decode_srv(consensus.shared_rand_current)
        else:
            # Fallback: use zero SRV (not ideal but allows testing)
            self.shared_random = b"\x00" * 32

        # Build the hashring
        self.entries: list[HSDirectoryEntry] = []
        self._build_ring(ed25519_map)

    def _decode_srv(self, srv: tuple[int, str]) -> bytes:
        """Decode shared random value from consensus format.

        Args:
            srv: Tuple of (num_reveals, base64_value)

        Returns:
            32-byte shared random value
        """
        _, value_b64 = srv
        return base64.b64decode(value_b64)

    def _build_ring(self, ed25519_map: dict[str, bytes] | None = None) -> None:
        """Build the HSDir hashring from consensus routers.

        Args:
            ed25519_map: Pre-computed map of microdesc_hash -> Ed25519 identity (optional).
                         If not provided, microdescriptors will be fetched.
        """
        # First, collect all HSDir routers and their microdesc hashes
        hsdir_routers: list[RouterStatusEntry] = []
        microdesc_hashes: list[str] = []

        for router in self.consensus.routers:
            if "HSDir" in router.flags and router.microdesc_hash:
                hsdir_routers.append(router)
                microdesc_hashes.append(router.microdesc_hash)

        # Use provided ed25519_map or fetch microdescriptors
        if ed25519_map is None:
            ed25519_map = self._fetch_ed25519_identities(microdesc_hashes)

        # Build the ring
        for router in hsdir_routers:
            # microdesc_hash is guaranteed non-None from the filter above
            md_hash = router.microdesc_hash
            if md_hash is None:
                continue
            ed25519_id = ed25519_map.get(md_hash)
            if ed25519_id is None:
                continue

            # Compute hsdir_index
            index = self._compute_hsdir_index(ed25519_id)

            entry = HSDirectoryEntry(
                router=router,
                index=index,
                ed25519_id=ed25519_id,
            )
            self.entries.append(entry)

        # Sort by index (position on ring)
        self.entries.sort()

    def _fetch_ed25519_identities(self, microdesc_hashes: list[str]) -> dict[str, bytes]:
        """Fetch Ed25519 identities from microdescriptors.

        Args:
            microdesc_hashes: List of microdescriptor hashes

        Returns:
            Dict mapping microdesc hash to 32-byte Ed25519 identity
        """
        # pylint: disable=import-outside-toplevel
        import time

        from torscope.cache import get_microdescriptor, save_microdescriptors
        from torscope.directory.authority import get_shuffled_authorities
        from torscope.directory.client import DirectoryClient
        from torscope.directory.microdescriptor import MicrodescriptorParser

        # pylint: enable=import-outside-toplevel

        result: dict[str, bytes] = {}

        # First check cache for already fetched microdescriptors
        missing_hashes: list[str] = []
        for h in microdesc_hashes:
            md = get_microdescriptor(h)
            if md and md.ed25519_identity:
                try:
                    b64 = md.ed25519_identity
                    padding = 4 - len(b64) % 4
                    if padding != 4:
                        b64 += "=" * padding
                    result[h] = base64.b64decode(b64)
                except (ValueError, AttributeError):
                    missing_hashes.append(h)
            else:
                missing_hashes.append(h)

        if not missing_hashes:
            return result

        # Fetch missing microdescriptors in larger batches for efficiency
        batch_size = 500  # Larger batches = fewer HTTP requests
        client = DirectoryClient()
        total_batches = (len(missing_hashes) + batch_size - 1) // batch_size
        authorities = get_shuffled_authorities()
        max_retries = 3

        print(f"Fetching {len(missing_hashes)} microdescriptors in {total_batches} batches...")

        for batch_num, i in enumerate(range(0, len(missing_hashes), batch_size)):
            batch = missing_hashes[i : i + batch_size]
            print(f"  Batch {batch_num + 1}/{total_batches}...", end="", flush=True)

            # Try with retries and different authorities
            success = False
            for retry in range(max_retries):
                # Rotate through authorities on retry
                authority = authorities[(batch_num + retry) % len(authorities)]
                try:
                    fetch_result = client.fetch_microdescriptors(batch, authority=authority)
                    if fetch_result:
                        content, _ = fetch_result
                        microdescriptors = MicrodescriptorParser.parse(content.decode("utf-8"))

                        # Save to cache and extract identities
                        save_microdescriptors(microdescriptors, authority.nickname, "authority")

                        for md in microdescriptors:
                            if md.ed25519_identity:
                                try:
                                    b64 = md.ed25519_identity
                                    padding = 4 - len(b64) % 4
                                    if padding != 4:
                                        b64 += "=" * padding
                                    result[md.digest] = base64.b64decode(b64)
                                except (ValueError, AttributeError):
                                    pass
                        print(f" OK ({len(microdescriptors)})")
                        success = True
                        break
                except (ConnectionError, OSError, TimeoutError):
                    if retry < max_retries - 1:
                        # Brief delay before retry
                        time.sleep(0.5 * (retry + 1))
                        continue

            if not success:
                print(" failed (after retries)")

        return result

    def _get_ed25519_id(self, router: RouterStatusEntry) -> bytes | None:
        """Get Ed25519 identity for a router.

        Uses the Ed25519 identity from the consensus "id ed25519" line if available.

        Args:
            router: Router status entry

        Returns:
            32-byte Ed25519 identity or None
        """
        if router.ed25519_identity is None:
            return None

        try:
            # The consensus provides base64-encoded Ed25519 identity
            # Add padding if needed (Tor uses unpadded base64)
            b64 = router.ed25519_identity
            padding = 4 - len(b64) % 4
            if padding != 4:
                b64 += "=" * padding
            return base64.b64decode(b64)
        except (ValueError, AttributeError):
            return None

    def _compute_hsdir_index(self, ed25519_id: bytes) -> bytes:
        """Compute HSDir index on the hashring.

        hsdir_index = H("node-idx" | ed25519_id | shared_random |
                        INT_8(period_num) | INT_8(period_length))

        Args:
            ed25519_id: 32-byte Ed25519 identity

        Returns:
            32-byte hash (position on ring)
        """
        data = (
            b"node-idx"
            + ed25519_id
            + self.shared_random
            + struct.pack(">Q", self.time_period)
            + struct.pack(">Q", self.period_length)
        )
        return sha3_256(data)

    def get_responsible_hsdirs(
        self,
        blinded_key: bytes,
        n_replicas: int = HS_N_REPLICAS,
        spread: int = HSDIR_SPREAD_FETCH,
    ) -> list[RouterStatusEntry]:
        """Find HSDirs responsible for a hidden service descriptor.

        For each replica, we compute the service's position on the ring
        and find the closest HSDirs.

        Args:
            blinded_key: 32-byte blinded public key of the hidden service
            n_replicas: Number of replicas (default: 2)
            spread: Number of HSDirs per replica (default: 3 for fetch)

        Returns:
            List of responsible HSDir routers (deduplicated)
        """
        if not self.entries:
            return []

        responsible: list[RouterStatusEntry] = []
        seen_fingerprints: set[str] = set()

        for replica in range(1, n_replicas + 1):  # 1-indexed per spec
            # Compute service's position for this replica
            hs_index = self._compute_hs_index(blinded_key, replica)

            # Find closest HSDirs on the ring
            closest = self._find_closest(hs_index, spread)

            for entry in closest:
                fp = entry.router.fingerprint
                if fp not in seen_fingerprints:
                    seen_fingerprints.add(fp)
                    responsible.append(entry.router)

        return responsible

    def _compute_hs_index(self, blinded_key: bytes, replica: int) -> bytes:
        """Compute hidden service index on the hashring.

        hs_index = H("store-at-idx" | blinded_key | INT_8(replica) |
                     INT_8(period_length) | INT_8(period_num))

        Note: INT_8 means 8 bytes (64-bit big-endian integer).

        Args:
            blinded_key: 32-byte blinded public key
            replica: Replica number (0-based)

        Returns:
            32-byte hash (position on ring)
        """
        data = (
            b"store-at-idx"
            + blinded_key
            + struct.pack(">Q", replica)
            + struct.pack(">Q", self.period_length)
            + struct.pack(">Q", self.time_period)
        )
        return sha3_256(data)

    def _find_closest(self, target: bytes, n: int) -> list[HSDirectoryEntry]:
        """Find n closest HSDirs to target position on the ring.

        Uses binary search to find insertion point, then takes n entries
        starting from that point (wrapping around).

        Args:
            target: Target position (32-byte hash)
            n: Number of entries to return

        Returns:
            List of closest HSDir entries
        """
        if not self.entries:
            return []

        # Binary search for insertion point
        left, right = 0, len(self.entries)
        while left < right:
            mid = (left + right) // 2
            if self.entries[mid].index < target:
                left = mid + 1
            else:
                right = mid

        # Take n entries starting from insertion point (with wraparound)
        result: list[HSDirectoryEntry] = []
        for i in range(min(n, len(self.entries))):
            idx = (left + i) % len(self.entries)
            result.append(self.entries[idx])

        return result

    @property
    def size(self) -> int:
        """Number of HSDirs on the ring."""
        return len(self.entries)

    @classmethod
    def fetch_ed25519_map(cls, consensus: ConsensusDocument) -> dict[str, bytes]:
        """Fetch Ed25519 identities for all HSDir relays.

        This is a convenience method to pre-fetch the Ed25519 map for use with
        multiple HSDir ring instances.

        Args:
            consensus: Network consensus document

        Returns:
            Dict mapping microdesc_hash -> 32-byte Ed25519 identity
        """
        # Collect all HSDir microdesc hashes
        microdesc_hashes: list[str] = []
        for router in consensus.routers:
            if "HSDir" in router.flags and router.microdesc_hash:
                microdesc_hashes.append(router.microdesc_hash)

        # Create temporary instance to use the fetch method
        dummy = object.__new__(cls)
        return dummy._fetch_ed25519_identities(microdesc_hashes)


def get_responsible_hsdirs(
    consensus: ConsensusDocument,
    blinded_key: bytes,
    time_period: int,
    n_replicas: int = HS_N_REPLICAS,
    spread: int = HSDIR_SPREAD_FETCH,
) -> list[RouterStatusEntry]:
    """Convenience function to find responsible HSDirs.

    Args:
        consensus: Network consensus
        blinded_key: 32-byte blinded public key
        time_period: Current time period number
        n_replicas: Number of replicas
        spread: Number of HSDirs per replica

    Returns:
        List of responsible HSDir routers
    """
    ring = HSDirectoryRing(consensus, time_period)
    return ring.get_responsible_hsdirs(blinded_key, n_replicas, spread)



"""V3 hidden service client authorization.

This module implements client authorization for v3 hidden services as specified
in rend-spec-v3.txt section 2.5.1.4.

Client authorization allows hidden service operators to restrict access to
specific clients who possess the correct x25519 private key.

Key format (standard Tor format):
    descriptor:x25519:<base32-encoded-private-key>

The client key file is typically stored at:
    ~/.tor/onion_auth/<onion-address>.auth_private
"""

from __future__ import annotations

import base64
from dataclasses import dataclass, field

from cryptography.hazmat.primitives.asymmetric.x25519 import (
    X25519PrivateKey,
    X25519PublicKey,
)
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from torscope.crypto import shake256


@dataclass
class AuthClientEntry:
    """A single auth-client entry from the first layer."""

    client_id: bytes  # 8 bytes
    iv: bytes  # 16 bytes
    encrypted_cookie: bytes  # 32 bytes


@dataclass
class ClientAuthData:
    """Parsed client auth data from the first layer plaintext."""

    auth_type: str  # "x25519"
    ephemeral_key: bytes  # 32-byte X25519 public key
    auth_clients: list[AuthClientEntry] = field(default_factory=list)


def parse_client_auth_key(auth_key_str: str) -> bytes:
    """Parse a client authorization key string.

    Accepts formats:
        - "<address>:descriptor:x25519:<base32-key>"  (full Tor file format)
        - "descriptor:x25519:<base32-key>"            (standard Tor format)
        - "<base32-key>"                              (just the key)

    Args:
        auth_key_str: The auth key string

    Returns:
        32-byte X25519 private key

    Raises:
        ValueError: If the key format is invalid
    """
    key_str = auth_key_str.strip()

    # Handle full Tor file format: <address>:descriptor:x25519:<key>
    if ":descriptor:x25519:" in key_str.lower():
        parts = key_str.split(":")
        if len(parts) >= 4:
            # Last part is the key
            key_str = parts[-1]
    # Handle standard Tor format: descriptor:x25519:<key>
    elif key_str.lower().startswith("descriptor:x25519:"):
        key_str = key_str[18:]  # Remove prefix

    # Decode base32 (Tor uses RFC 4648 without padding)
    try:
        # Add padding for standard base32 decoder
        padded = key_str.upper() + "=" * ((8 - len(key_str) % 8) % 8)
        key_bytes = base64.b32decode(padded)
    except Exception as e:
        raise ValueError(f"Invalid base32 encoding in auth key: {e}") from e

    if len(key_bytes) != 32:
        raise ValueError(f"Auth key must be 32 bytes, got {len(key_bytes)}")

    return key_bytes


def read_client_auth_file(file_path: str) -> bytes:
    """Read client authorization key from a file.

    The file should contain a line in one of these formats:
        - <address>:descriptor:x25519:<base32-key>  (standard Tor format)
        - descriptor:x25519:<base32-key>
        - <base32-key>

    The file is typically stored at:
        ~/.tor/onion_auth/<onion-address>.auth_private

    Args:
        file_path: Path to the auth key file

    Returns:
        32-byte X25519 private key

    Raises:
        ValueError: If the file format is invalid
        FileNotFoundError: If the file doesn't exist
    """
    with open(file_path, encoding="utf-8") as f:
        content = f.read().strip()

    # Skip empty lines and comments, take first valid line
    for line in content.split("\n"):
        line = line.strip()
        if line and not line.startswith("#"):
            return parse_client_auth_key(line)

    raise ValueError(f"No valid auth key found in {file_path}")


def parse_client_auth_data(first_layer_text: str) -> ClientAuthData | None:
    """Parse client auth data from first layer plaintext.

    Extracts:
        - desc-auth-type (must be "x25519")
        - desc-auth-ephemeral-key (32-byte X25519 pubkey)
        - auth-client entries (client-id, iv, encrypted-cookie)

    Args:
        first_layer_text: Decrypted first layer as text

    Returns:
        ClientAuthData if auth is present, None if no auth required
    """
    lines = first_layer_text.strip().split("\n")

    auth_type: str | None = None
    ephemeral_key: bytes | None = None
    auth_clients: list[AuthClientEntry] = []

    for line in lines:
        line = line.strip()

        if line.startswith("desc-auth-type "):
            auth_type = line.split()[1]

        elif line.startswith("desc-auth-ephemeral-key "):
            key_b64 = line.split()[1]
            ephemeral_key = _decode_base64(key_b64)

        elif line.startswith("auth-client "):
            parts = line.split()
            if len(parts) >= 4:
                client_id = _decode_base64(parts[1])
                iv = _decode_base64(parts[2])
                encrypted_cookie = _decode_base64(parts[3])
                auth_clients.append(
                    AuthClientEntry(
                        client_id=client_id,
                        iv=iv,
                        encrypted_cookie=encrypted_cookie,
                    )
                )

    # If no auth data found, return None
    if auth_type is None:
        return None

    if auth_type != "x25519":
        raise ValueError(f"Unsupported auth type: {auth_type}")

    if ephemeral_key is None:
        raise ValueError("Missing desc-auth-ephemeral-key")

    if len(ephemeral_key) != 32:
        raise ValueError(f"Invalid ephemeral key length: {len(ephemeral_key)}")

    return ClientAuthData(
        auth_type=auth_type,
        ephemeral_key=ephemeral_key,
        auth_clients=auth_clients,
    )


def derive_client_id_and_cookie_key(
    client_privkey: bytes,
    ephemeral_pubkey: bytes,
    subcredential: bytes,
) -> tuple[bytes, bytes]:
    """Derive CLIENT-ID and COOKIE-KEY for client authorization.

    Per rend-spec-v3:
        SECRET_SEED = x25519(client_privkey, ephemeral_pubkey)
        KEYS = KDF(subcredential | SECRET_SEED, 40)
        CLIENT-ID = KEYS[0:8]
        COOKIE-KEY = KEYS[8:40]

    Args:
        client_privkey: 32-byte X25519 private key
        ephemeral_pubkey: 32-byte X25519 public key from descriptor
        subcredential: 32-byte subcredential

    Returns:
        Tuple of (client_id, cookie_key)
    """
    # Load keys
    privkey = X25519PrivateKey.from_private_bytes(client_privkey)
    pubkey = X25519PublicKey.from_public_bytes(ephemeral_pubkey)

    # Compute shared secret
    secret_seed = privkey.exchange(pubkey)

    # Derive keys: KDF(subcredential | SECRET_SEED, 40)
    kdf_input = subcredential + secret_seed
    keys = shake256(kdf_input, 40)

    client_id = keys[:8]
    cookie_key = keys[8:]

    return client_id, cookie_key


def decrypt_descriptor_cookie(
    auth_clients: list[AuthClientEntry],
    client_id: bytes,
    cookie_key: bytes,
) -> bytes | None:
    """Find matching auth-client entry and decrypt the descriptor cookie.

    Args:
        auth_clients: List of auth-client entries from descriptor
        client_id: 8-byte client identifier
        cookie_key: 32-byte key for decryption

    Returns:
        32-byte descriptor_cookie or None if no matching entry
    """
    # Find matching auth-client entry
    for entry in auth_clients:
        if entry.client_id == client_id:
            # Decrypt using AES-256-CTR
            # descriptor_cookie = AES-CTR(cookie_key, iv) XOR encrypted_cookie
            cipher = Cipher(algorithms.AES(cookie_key), modes.CTR(entry.iv))
            decryptor = cipher.decryptor()
            descriptor_cookie = decryptor.update(entry.encrypted_cookie) + decryptor.finalize()
            return descriptor_cookie

    return None


def get_descriptor_cookie(
    first_layer_text: str,
    client_privkey: bytes,
    subcredential: bytes,
) -> bytes | None:
    """Get descriptor cookie for a private hidden service.

    This is the main entry point for client authorization.

    Args:
        first_layer_text: Decrypted first layer plaintext
        client_privkey: 32-byte X25519 private key from auth file
        subcredential: 32-byte subcredential

    Returns:
        32-byte descriptor_cookie, or None if:
        - No client auth required (public service with no auth entries)
        - No matching auth-client entry found (service may still allow public access)

    Raises:
        ValueError: If auth data is malformed
    """
    # Parse auth data from first layer
    auth_data = parse_client_auth_data(first_layer_text)

    if auth_data is None:
        # No client auth required (public service)
        return None

    # Derive our client-id and cookie-key
    client_id, cookie_key = derive_client_id_and_cookie_key(
        client_privkey=client_privkey,
        ephemeral_pubkey=auth_data.ephemeral_key,
        subcredential=subcredential,
    )

    # Find and decrypt our cookie
    descriptor_cookie = decrypt_descriptor_cookie(
        auth_clients=auth_data.auth_clients,
        client_id=client_id,
        cookie_key=cookie_key,
    )

    # Return the cookie (or None if no match - caller will try public decryption)
    return descriptor_cookie


def _decode_base64(s: str) -> bytes:
    """Decode base64 with padding handling."""
    # Add padding if needed
    padding = 4 - len(s) % 4
    if padding != 4:
        s += "=" * padding
    return base64.b64decode(s)
